#!/usr/bin/env python

"""Drawing functions for constructing Marks in the toyplot context.

"""

# main tree drawing Mark and .draw() function
from toytree.drawing.src.mark_toytree import ToyTreeMark
from toytree.drawing.src.draw_toytree import draw_toytree

# import custom render modules to expose multidispatch calls
import toytree.drawing.src.render_tree
import toytree.drawing.src.render_annotation
import toytree.drawing.src.render_pie
# from toytree.drawing.src.render_text import render_text

# canvas/axes setup
# from toytree.drawing.src.setup_canvas import get_canvas_and_axes
# from toytree.drawing.src.setup_grid import Grid  #Setup



#!/usr/bin/env python

"""A simple immutable Node class.

A Node represents a vertex in a tree or graph, with connections
represented by pointers to other Node objects in its `.up` and
`.children` attributes. The core purpose of Node objects it to
provide *traversal* functions for accessing connected Nodes in an
order determined by their connections.

Notes
-----
This class is inspired by the ete3.TreeNode class object, and retains
similarities in syntax to make it easy to learn for ete users. However,
it is greatly simplified, and differs in several other important
respects. Most notably, the toytree.Node class is intended to be
_immutable_, meaning users are not able to directly edit .dist, .up,
.children, or other attributes of Node objects that affect the tree
topology. Instead, ToyTree class objects have functions for modifying
Nodes that do so in the context of updating the entire tree (i.e.,
updating other Node's attributes if needed). Thus, toytree.Node is a
more minimal object used mainly for traversal and to store Node data.

References
----------
- https://en.wikipedia.org/wiki/Tree_traversal
- https://github.com/ete3
"""

# pylint: disable=no-name-in-module
# pylint: disable=invalid-name
# pylint: disable=too-many-public-methods
# pylint: disable=no-self-use, unused-argument


from __future__ import annotations
from typing import List, Optional, Union, Iterator, Tuple  # Set, Any
from functools import total_ordering
from copy import deepcopy
from collections import deque

from loguru import logger
import numpy as np
from toytree.utils import TreeNodeError

# register the logger
logger = logger.bind(name="toytree")


class Node:
    """Node class representing a single vertex in a ToyTree.

    Node objects store the parent/child relationships among directly
    connected Nodes, as well as attributes/features of a Node. The
    core purpose of Node objects it to provide *traversal* functions
    for accessing connected Nodes in an order determined by their
    connections. The root of a set of connected Nodes is stored as
    the `treenode` attribute of a ToyTree class object.

    Note
    ----
    Node class objects are intended to be *mostly* immutable, i.e.,
    not directly modified by users. This is to reduce user errors by
    modifying Node features that affect the tree topology or values.
    Nodes are mutable, however, in terms of assigning and modifying
    arbritrary data to Nodes (e.g., `tree[3].body_size = 10`).

    Although Nodes *can* be modified directly, the recommended way
    is to use specific ToyTree class functions to change attributes
    of Nodes (e.g., change topology, edge lengths, rotations, rooting,
    features). See functions in `ToyTree.mod` to modify the tree, and
    see `ToyTree.set_node_data` to assign or modify data on Nodes.

    Parameters
    ----------
    name: str
        Node name is shown as either a tip or node label in tree
        drawings, and can be any string, including HTML.
    dist: float
        Node dist value is the edge length from this Node to its parent.
    support: float
        Node support value for the edge from this Node to its parent.

    Examples
    --------
    >>> # you can create a Node directly, but this is not intended.
    >>> node = toytree.Node(name='root')
    >>>
    >>> # instead, create a ToyTree and access Nodes from it.
    >>> tree = toytree.rtree.unittree(ntips=10, seed=123)
    >>>
    >>> # Nodes can be accessed by traversal from ToyTrees or Nodes
    >>> for node in tree.traverse("postorder"):
    >>>     print(node)
    >>>
    >>> # or, Nodes can be indexed from ToyTrees by their idx label.
    >>> node = tree[3]
    >>>
    >>> # other Nodes can be accessed relative to this one by traversal
    >>> parent = node.up
    >>> ancs = node.get_ancestors()
    >>> descs = node.get_descendants()
    >>>
    >>> # Node data/attributes can be accessed from a Node
    >>> print(node.name, node.idx, node.dist)
    >>>
    >>> # but cannot be modified on Nodes (they are immutable)
    >>> node.dist = 10  # raises a TreeNodeError
    """

    def __init__(self, name: str = "", dist: float = 0.0, support: float = np.nan):
        self._name = str(name)
        """: name string assigned to Node."""
        self._dist = float(dist)
        """: length value associated to the edge above this Node."""
        self._support = float(support)
        """: support value associated to the edge above this Node."""

        # Non-init attributes.
        self._children: Tuple[Node] = tuple()
        """: tuple of Node instances directly descended from this Node."""
        self._up: Optional[Node] = None
        """: the ancestor Node instance (towards root) from this Node."""
        self._idx: int = -1
        """: unique node index label (idx) assigned to Nodes in a ToyTree."""
        self._height: float = 0.0
        """: height of this Node above the connected Node farthest from root."""
        self._x: float = 0.0
        """: private attribute updated during drawing as x-coordinate."""

    @property
    def name(self) -> str:
        """Name string assigned to Node."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Set the 'name' attribute, forced as a string."""
        self._name = str(value)

    @property
    def dist(self) -> float:
        """Edge length associated to the edge above this Node."""
        return self._dist

    @dist.setter
    def dist(self, value: float) -> None:
        """Set the dist attribute, forced as an int or float."""
        raise TreeNodeError(
            "Cannot set .dist attribute of a Node. If you are an "
            "advanced user then you can do so by setting ._dist. "
            "See the docs section on Modifying Nodes and Tree Topology."
        )

    @property
    def support(self) -> float:
        """Return the support value for the edge above this Node."""
        return self._support

    @support.setter
    def support(self, value: float) -> None:
        """Set a 'support' attribute. This is like any other feature
        that can be set on a Node, but defaults to nan for all Nodes."""
        try:
            self._support = float(value)
        except ValueError as err:
            raise TreeNodeError("Node support must be int or float") from err

    @property
    def height(self) -> float:
        """Return the cached height of this Node.

        Node heights are calculated relative to a connected Node
        that is furthest from the root Node, defined as height 0.

        `height` is an emergent property of a collection of Nodes, and
        thus cannot be set for an individual Node. It can, however, be
        set from a ToyTree (collection of Nodes), using `set_node_data`,
        which will modify the `.dist` attributes of multiple Nodes to
        set the height of one or more Nodes to the entered values. This
        can also be set using `toytree.mod.set_node_heights()`.
        """
        return self._height

    @height.setter
    def height(self, value: float) -> None:
        """height Node attribute cannot be set by user.        """
        raise TreeNodeError(
            "Cannot set .height attribute of a Node since it is an emergent "
            "property of multiple Nodes. The prefered method of setting Node "
            "heights is to use `set_node_data` from a ToyTree, or "
            "`set_node_heights` from the `.mod` submodule. See the docs "
            "section on Modifying Nodes and Tree Topology for more details. "
            "Finally, if you know what you are doing you *can* force the "
            "height of a Node by setting its `._height` attribute."
        )

    @property
    def children(self) -> Tuple[Node]:
        """Return a list of child Nodes of this Node."""
        return self._children

    @children.setter
    def children(self, value: Node) -> None:
        """The children tuple cannot be modified. See ToyTree.mod."""
        raise TreeNodeError(
            "Cannot set .children attribute of a Node. If you are an "
            "advanced user then you can do so by setting ._children. "
            "See the docs section on Modifying Nodes and Tree Topology."
        )

    @property
    def up(self) -> Union[Node, None]:
        """The parent node (next node towards root) from this node."""
        return self._up

    @up.setter
    def up(self, value: Node) -> None:
        """Modifying up requires a call to ToyTree._update()."""
        raise TreeNodeError(
            "Cannot set .up attribute of a Node. If you are an "
            "advanced user then you can do so by setting ._up. "
            "See the docs section on Modifying Nodes and Tree Topology."
        )

    @property
    def idx(self) -> int:
        """Return the unique integer idx label of this Node."""
        return self._idx

    @idx.setter
    def idx(self, value: int):
        """This will raise an TreeNodeError, users cannot set idx."""
        raise TreeNodeError(
            "Cannot set .idx attribute of a Node. If you are an "
            "advanced user then you can do so by setting ._idx. "
            "See the docs section on Modifying Nodes and Tree Topology."
        )

    ################################################
    # IDENTITY
    ################################################

    def is_leaf(self) -> bool:
        """Return True if Node is a leaf (terminal)."""
        return not bool(self.children)

    def is_root(self):
        """Return True if Node has no parent."""
        return self.up is None

    def copy(self, detach: bool = False) -> Node:
        r"""Return a deepcopy of this Node (and its connected Nodes).

        All connected Nodes (ancestral and descendant) are also copied
        and the can be referenced from the returned Node. The Node
        that calls .copy() can optionally be detached from its
        ancestors and returned as the root of its subtree. This is
        slightly faster since only descendant Nodes are copied.

                4   when .copy called from Node 3   n4
               / \    all connected Nodes are       / \
              3   \     also copied ------>       n3   \
             / \   \                              / \   \
            0   1   2                           n0  n1  n2

        Parameters
        ----------
        detach: bool
            If True the Node is returned as the treenode (root) of its
            subtree with all Nodes of the subtree copied. If False any
            ancestral Nodes are also copied and are connected to the
            returned Node.

        See Also
        --------
        `Toytree.copy()`, `toytree.mod.extract_subtree`

        Examples
        --------
        >>> tree = toytree.tree("((a,b),c);")
        >>> node = tree.get_mrca_node("a", "b")
        >>> cnode = node.copy(detach=True)
        >>> subtree = toytree.tree(cnode)
        """
        # return a deepcopy of this Node and its descendants detached
        # from any ancestor Nodes.
        if detach:
            # save connection to ancestors and remove it from this Node
            saved_connection = self.up
            self._up = None
            # create deepcopy of this Node and descendants (not ancestors)
            copied_node = deepcopy(self)
            # restore connection to ancestors for the original Node
            self._up = saved_connection
            return copied_node

        # get connected root Node
        treenode = self.get_treenode()

        # make a deepcopy of root, otherwise only nested are copied.
        # TODO: is it still necessary to do this?
        copied_treenode = deepcopy(treenode)

        # find focal node to be returned
        for node in copied_treenode.traverse():
            if (node.idx == self.idx) and (node.name == self.name):
                return node
        raise TreeNodeError("copy failed, tree structure is broken.")

    #################################################
    # DUNDERS
    #################################################

    def __repr__(self) -> str:
        """Returned string showing Node idx and name only if present."""
        _idx = "" if self.idx == -1 else f"idx={self.idx}"
        _name = "" if not self.name else f"name='{self.name}'"
        inner = ", ".join([i for i in (_idx, _name) if i])
        return f"<Node({inner})>"

    # def __str__(self) -> str:
    #     """Printed string representation of Node"""
    #     return self.__repr__()f"Node(idx={self.idx}, {pic})"

    def __len__(self) -> int:
        """Return length of Node as number of descendant leaf Nodes"""
        return sum(1 for i in self.iter_leaves())

    @total_ordering  # auto-creates all other ordering operations
    def __gt__(self, other: Node) -> bool:
        """Return True if this Node's idx is greater than the others."""
        if self._idx > other._idx:
            return True
        return False

    def __hash__(self) -> int:
        """Return a hash of the Node based on its repr."""
        return hash(repr(self))

    #####################################################
    # NODE CONNECTIONS
    # after modifying connections among Nodes in a ToyTree the
    # ToyTree._update() function must be called. See the
    # ToyTree.mod subpackage for alternative tree modification funcs.
    # These funcs are primarily for internal/developer use.
    #####################################################

    def _add_child(self, node: Node) -> None:
        """Connect a Node by adding it as a child to this one.

        See the ToyTree.mod subpackage for alternative user-facing
        functions for adding Nodes to a tree.

        Note
        ----
        If you use this private function to add connections among Nodes
        be aware that you are modifying the tree, and thus need to
        follow `Node/Tree updating rules in toytree <toytree.rtfd.io>`_.
        """
        node._up = self
        self._children += (node,)

    def _remove_child(self, node: Node) -> None:
        """Remove a connection from this Node to a child Node."""
        # should we add assertion that children do not have self as .up?
        self._children = tuple(i for i in self._children if i != node)

    def _delete(self, preserve_dists: bool = True, prevent_unary: bool = False) -> None:
        r"""Delete a Node from a tree.

        This removes a Node object and its connections to other Nodes.
        The `.up` and `.children` attributes of any connected Nodes are
        modified.

                    4                         4
                   / \       delete 3        /|\
                  3   \      ------->       / | \
                 / \   \                   /  |  \
                0   1   2                 0   1   2

        Parameters
        ----------
        preserve_dists: bool
            If True then children inherit the dist values of their
            deleted parents so that their dist value reaches their
            grandparents original height. If False, children retain
            their original dist but are connected to their grandparent.
        prevent_unary: bool
            If True then any unary Nodes (internal nodes with a single
            child) that are left behind from the deletion process will
            be removed as well. For example, if Node 1 is deleted in
            the example above then Node 3 would also be deleted to
            connect Node 0 to Node 4.

        See Also
        --------
        `toytree.mod.remove_node`, `toytree.mod.remove_unary_nodes`

        Example
        --------
        >>> tree = toytree.tree("((a,b),c);")
        >>> node = tree.get_mrca_node('a', 'b')
        >>> node._delete()
        >>> tree._update()  # required b/c Nodes changed
        >>> tree.draw()
        """
        # get parent node
        grandparent = self.up
        if not grandparent:
            logger.warning("cannot delete root Node.")
            return

        # each child inherits deleted Node's dist
        if preserve_dists:
            for child in self._children:
                child._dist += self._dist

        # connect children to grandparent and rm reference to self
        for child in self._children:
            grandparent._add_child(child)
        grandparent._remove_child(self)

        # do not allow parents with only one child
        if prevent_unary and len(grandparent.children) < 2:
            grandparent._delete(preserve_dists, prevent_unary)

        # Node cannot be deleted here, but all connections are severed.
        self._idx = -1
        self._up = None
        self._children = ()


    def _detach(self) -> Node:
        r"""Return this Node as a subtree detached from its ancestor.

        This Node and its descendants are detached from their ancestor
        Nodes. In other words, this removes the parent from this Node's
        `up` and removes this Node from its parent's `children`. This
        does not create a copy of any Nodes (see Node.copy()), thus
        other Nodes connected to this one are also affected. In the
        example below .detach() is called on Node 3 and Node 3 is
        returned as a subtree of the original tree structure, but note
        that Node 4 was also modified.

                    4                                   4
                   / \       detach 3       |           |
                  /   3      ------->       3           |
                 /   / \                   / \          |
                0   1   2                 1   2         0
                                        [Returned]  [Modified]

        Note
        ----
        The Node will preserve its .dist (edge length) attribute, and
        can be connected to another Node using the `Node._add_child`
        method. The detach method is primarily used internally.

        See Also
        --------
        `Node.copy`, `toytree.mod.extract_subtree`

        Example
        -------
        >>> tree = toytree.tree("((1,2)3,0)4;")
        >>> # get (1,2) detached as subtree
        >>> node = tree.get_mrca_node('1', '2')
        >>> subtree_node = node._detach()
        """
        if self._up:
            self._up._children = self.get_sisters()
            self._up = None
        return self

    #################################################
    # TRAVERSAL                                    ##
    #################################################

    def traverse(self, strategy: str = "levelorder") -> Iterator[Node]:
        """Visit all connected Nodes using a tree traversal strategy.

        Notes
        -----
        preorder:
            Parents are visited before children. Traverses all the way
            down each left subtree before proceeding to right child.
        postorder:
            Children are visited before parents. The left subtree is
            visited, then right, then the parent.
        levelorder:
            Nodes the same distance from root are visited left to
            right, before descending to next level.
        idxorder:
            Leaf nodes are visited left to right, followed by internal
            nodes in postorder traversal order.
        inorder:
            Nodes are visited in non-descreasing order if they are
            a binary search tree: left child, parent, right child.

        Parameters
        ----------
        strategy: str
            A traversal strategy for the order in which nodes will
            be visited: 'preorder', 'postorder', 'levelorder',
            'inorder', or 'idxorder'.

        Examples
        --------
        >>> tree = toytree.rtree.unittree(10)
        >>> for order, node in enumerate(tree.treenode.traverse("postorder")):
        >>>     node.name = str(order)
        >>> tree.draw(node_labels="name", node_sizes=18);
        """
        if strategy == "preorder":
            return self._traverse_preorder()
        if strategy == "levelorder":
            return self._traverse_levelorder()
        if strategy == "postorder":
            return self._traverse_postorder()
        if strategy == "inorder":
            return self._traverse_inorder()
        if strategy == "idxorder":
            return self._traverse_idxorder()
        raise TreeNodeError(
            "supported strategies are ['idxorder', 'preorder', "
            "'postorder', 'levelorder', and 'inorder']"
        )

    def _traverse_idxorder(self) -> Iterator[Node]:
        """Iterates over all nodes in 'idx' order.

        This is a non-standard tree traversal algorithm that is used
        for working with phylogenetic trees, where the leaf nodes
        are typically of special interest as the extant samples, and
        thus numbering them sequentially (0-ntips) is convenient. All
        internal nodes are labeled by post-order from (ntips-nnodes).
        The order is topologically sorted.

        See Also
        ---------
        ToyTree._update
            A similar idxorder traversal is performed by ToyTree class
            objects in the private update function, which is called
            everytime a tree is init or topology changed to update the
            Node idx labels and cache the idxorder. This func also
            sets Node height and spacing during traversal.
        """
        queue = [self]
        inner_stack = []
        outer_stack = []
        while queue:

            # get node from end of the queue
            node = queue.pop()

            # if node is leaf insert to output stack ntips from right
            if node.is_leaf():
                outer_stack.append(node)

            # if not leaf then insert to output stack...
            else:
                inner_stack.append(node)

            # add node's children to the queue
            queue.extend(node.children)

        # return nodes in reverse order they were added to stack
        while outer_stack:
            yield outer_stack.pop()
        while inner_stack:
            yield inner_stack.pop()

    def _traverse_postorder(self) -> Iterator[Node]:
        """Iterate over all descendant nodes in tip-to-root order.

        This visits all children before parents by visiting nodes...
        """
        queue = [self]
        stack = []
        while queue:

            # push node from queue onto the output stack
            node = queue.pop()
            stack.append(node)

            # push node's children to the queue
            queue.extend(node.children)

        # return nodes in reverse order they were added to stack
        while stack:
            yield stack.pop()

    def _traverse_preorder(self) -> Iterator[Node]:
        """Iterate over all nodes by 'preorder' traversal.

        This visits parents before children, by visiting all the way
        down the left and then right descendants of each node.
        """
        queue = [self]
        while queue:
            # select node from right end of queue
            node = queue.pop()
            yield node

            # add node's children to queue in [right, left] order.
            queue.extend(node.children[::-1])

    def _traverse_levelorder(self) -> Iterator[Node]:
        """Iterate over all descendant nodes in levelorder.

        This is also called breadth-first search (BFS). It starts at
        the root and visits all nodes that are the same distance from
        the root (number of nodes away) before visiting the next level
        of nodes.
        """
        queue = deque([self])
        while queue:
            node = queue.popleft()
            yield node
            queue.extend(node.children)

    def _traverse_inorder(self) -> Iterator[Node]:
        """Iterate over all nodes by 'inorder' traversal.

        This algorithm traverse up each left subtree before each right
        subtree, and is intended for use on binary-search-trees. It
        may not give intended results when trees are non-binary.
        """
        queue = []
        node = self

        # if both the queue and current node are empty traversal is done
        while queue or node:

            # if current node exists then store it until later
            # and select its left child
            if node:
                queue.append(node)
                try:
                    node = node.children[0]
                except IndexError:
                    node = None

            # if it is empty then grab node from queue and select its
            # right child.
            else:
                node = queue.pop()
                yield node
                try:
                    node = node.children[1]
                except IndexError:
                    node = None

    #################################################
    # LEAF AND LEAF NAME RETRIEVAL                 ##
    # Defaults to returning in 'idxorder'          ##
    #################################################

    def iter_leaves(self) -> Iterator[Node]:
        """Return a Generator of leaves descended from this node in
        idxorder."""
        for node in self.traverse(strategy="idxorder"):
            if not node.children:
                yield node

    def get_leaves(self) -> List[Node]:
        """Return a list of leaf nodes descended from this node in
        idxorder."""
        return list(self.iter_leaves())

    def iter_leaf_names(self) -> Iterator[str]:
        """Return a Generator of names of Nodes descended from this
        node in idxorder."""
        for node in self.iter_leaves():
            yield node.name

    def get_leaf_names(self) -> List[str]:
        """Return a list of names of Nodes descended from this node
        in idxorder."""
        return list(self.iter_leaf_names())

    #################################################
    # NODE RELATIVE RETRIEVAL / TRAVERSAL
    #################################################

    def iter_sisters(self) -> Iterator[Node]:
        """Return a Generator to iterate over sister nodes."""
        if self._up:
            for child in self._up._children:
                if child != self:
                    yield child

    def get_sisters(self) -> Tuple[Node]:
        """Return list of other Nodes that are children of same parent."""
        return tuple(self.iter_sisters())

    def iter_descendants(self, strategy: str = "levelorder") -> Iterator[Node]:
        """A generator of descendant Nodes (including self)."""
        for node in self.traverse(strategy=strategy):
            yield node

    def get_descendants(self, strategy: str = "levelorder") -> Tuple[Node]:
        """Return a list of descendant Nodes (including self)."""
        return tuple(self.iter_descendants(strategy=strategy))

    def iter_ancestors(self, root: Optional[Node] = None, include_self: bool = False) -> Iterator[Node]:
        """Return a Generator of Nodes on path from this node to root.

        Parameters
        ----------
        root: Node or None
            Iteration will stop when <root> is reached, without yielding
            <root>. Default is None, which occurs above the tree root
            Node. You can optionally enter a Node to serve as the root/
            endpoint of iteration.
        include_self: bool
            If True then this Node will be returned as its own first
            ancestor, else its parent (.up) will be the first ancestor.

        Note
        ----
        'include_self' arg overrides 'root' and will return this Node
        if it is both self and root.
        """
        node = self

        # optionally yield self
        if include_self:
            yield self

        # end iteration if self == root
        if self == root:
            return

        # continue up tree yielding nodes until root or None is reached.
        node = node._up
        while 1:
            if node in (root, None):
                break
            yield node
            node = node._up

    def get_ancestors(self, root: Optional[Node] = None, include_self: bool = False) -> Tuple[Node]:
        """Return a tuple of Nodes on path from this node to root.

        Parameters
        ----------
        root: Node or None
            Iteration will stop when <root> is reached, without yielding
            <root>. Default is None, which occurs above the tree root
            Node. You can optionally enter a Node to serve as the root/
            endpoint of iteration.
        include_self: bool
            If True then this Node will be returned as its own first
            ancestor, else its parent (.up) will be the first ancestor.

        Note
        ----
        'include_self' arg overrides 'root' and will return this Node
        if it is both self and root.
        """
        return tuple(self.iter_ancestors(root=root, include_self=include_self))

    def get_treenode(self) -> Node:
        """Return the treenode (top Node) from a connected set of Nodes.

        This calls .up recursively until the top Node is reached and
        returned.

        Note
        ----
        It is much faster to fetch the treenode from a ToyTree object
        using its cached info at ToyTree.treenode.

        Examples
        --------
        >>> tree = toytree.tree("((a,b),c);")
        >>> t0 = tree[-1]
        >>> t1 = tree.treenode
        >>> t2 = tree[0].get_treenode()
        >>> assert t0 == t1 == t2
        """
        node = self
        while 1:
            if node.up:
                node = node.up
            else:
                return node


    #####################################################
    # TO TOYTREE                                        #
    # return a ToyTree with this Node as root treenode. #
    #####################################################
    # def to_toytree(self):
    #     """Return this Node as root treenode of a ToyTree class object."""
    #     return ToyTree(self.copy())

    # def _update_heights(self) -> None:
    #     """Sets the Node .height attribute by checking all other Nodes.

    #     This is intended for internal use only. It must be called to
    #     update Node heights if a tree topology or branch lengths have
    #     been modified. This will update *all* connected Node height
    #     values in-place.

    #     Note
    #     ----
    #     Height attributes are always updated when Nodes are modified
    #     by using built-in functions from ToyTree objects, such as in
    #     `toytree.mod`. This is why Node objects are purposefuly made
    #     to be immutable, to prevent users from changing the tree
    #     structure without properly updating connected Nodes, which
    #     could make height attributes incorrect.
    #     """
    #     # get the root node
    #     root = self
    #     while 1:
    #         if root.up:
    #             root = root.up
    #         else:
    #             break

    #     # get distance from each node to the root
    #     max_dist = 0.
    #     for node in root.traverse("preorder"):
    #         if node.up:
    #             node._height = node.dist + node.up._height
    #         else:
    #             node._height = 0
    #         if node.is_leaf():
    #             max_dist = max(max_dist, node._height)

    #     # set height as distance above Node farthest from the root.
    #     for node in root.traverse("preorder"):
    #         node._height = max_dist - node._height

    def _get_ascii(self, char1="-", compact=False):
        """Return the ASCII representation of a tree.

        Code based on the PyCogent GPL project.
        """
        if self.is_leaf():
            return ([char1 + "-" + self.name], 0)

        # internal nodes require spacing concerns
        space = max(3, len(self.name) if not self.children else 3)
        padding = " " * space
        pad = " " * (space - 1)
        mids = []
        result = []
        for child in self.children:
            if len(self.children) == 1:
                char2 = "/"
            elif child is self.children[0]:
                char2 = "/"
            elif child is self.children[-1]:
                char2 = "\\"
            else:
                char2 = "-"

            # recursion
            (clines, mid) = child._get_ascii(char2, compact=compact)
            mids.append(mid + len(result))
            result.extend(clines)
            if not compact:
                result.append("")
        if not compact:
            result.pop()
        (lo, hi, end) = (mids[0], mids[-1], len(result))

        prefixes = (
            [padding] * (lo + 1) + [pad + "|"] * (hi - lo - 1) + [padding] * (end - hi)
        )
        mid = int((lo + hi) / 2)
        prefixes[mid] = char1 + "-" * (space - 2) + prefixes[mid][-1]
        result = [x + y for (x, y) in zip(prefixes, result)]
        return (result, mid)

    def draw_ascii(self, compact: bool = False):
        """Return the ASCII drawing of a Node and its descendants.

        Code based on the PyCogent GPL project.
        """
        lines, _ = self._get_ascii(compact=compact)
        tree_lines = "\n".join(lines)
        print(f"\n{tree_lines}")


if __name__ == "__main__":

    nodes = {i: Node(name=i) for i in range(20)}
    for idx, n in enumerate(nodes):
        nodes[n]._idx = idx

    nodes[1]._add_child(nodes[2])
    nodes[1]._add_child(nodes[3])
    nodes[2]._add_child(nodes[4])
    nodes[3]._add_child(nodes[5])
    nodes[3]._add_child(nodes[6])
    nodes[5]._add_child(nodes[7])
    nodes[5]._add_child(nodes[8])

    for n in nodes[1]._traverse_levelorder():
        print(n.name)

    print("postorder")
    for n in nodes[1]._traverse_postorder():
        print(n.name)

    print("inorder")
    for n in nodes[1]._traverse_inorder():
        print(n.name)

    print("preorder")
    for n in nodes[1]._traverse_preorder():
        print(n.name)

    print("idxorder")
    for n in nodes[1]._traverse_idxorder():
        print(n.name)

    # print(nodes[3].copy())
    # print(nodes[3].__dict__.keys())

    import toytree

    tree = toytree.rtree.rtree(10)
    tree[3].hello = "3"
    print(tree.features)
    print(tree.get_node_data("support"))
    # print(tree.get_node_data('hello'))

    tree.treenode.draw_ascii(compact=False)

    # x = tree.copy()
    # print(x[0] <= x[1])



#!/usr/bin/env python

"""Random Tree generation functions. Uses numpy RNG.

"""

from typing import Optional
import numpy as np
import pandas as pd
from loguru import logger
from toytree.core.tree import ToyTree
from toytree.core.node import Node
from toytree.utils import ToytreeError

# pylint: disable=invalid-name, too-many-branches, too-many-statements

logger = logger.bind(name="toytree")


def _assign_names(tree: ToyTree, random: bool, rng: np.random.Generator) -> ToyTree:
    """Set name attributes on Nodes of a tree in place."""
    if not random:
        for nidx in range(tree.ntips):
            tree[nidx].name = f"r{nidx}"
    else:
        tidxs = rng.choice(range(tree.ntips), tree.ntips, replace=False)
        for nidx in range(tree.ntips):
            tree[nidx].name = f"r{tidxs[nidx]}"


def rtree(ntips: int, random_names: bool = False, seed: Optional[int] = None) -> ToyTree:
    """Return a random topology (fastest method).

    The default settings for Node support (np.nan) and dist (1.0) can
    be further modified after generating the tree.

    Parameters
    ----------
    ntips: int
        The number of tips in the tree.
    random_names: bool
        If True then tip names are randomized rather than ordered.
    seed: Optional[int]
        An integer seed for the numpy random number generator.

    Example
    -------
    >>> tree = toytree.rtree.rtree(ntips=10, seed=123)
    """
    # seed generator
    rng = np.random.default_rng(seed)

    # generate random topology with N tips.
    root = Node()
    current_tips = [root]

    # add children n-1 times, updating who the current children are.
    for _ in range(ntips - 1):
        tip = rng.choice(current_tips)
        current_tips.remove(tip)
        child0 = Node(dist=1.)
        child1 = Node(dist=1.)
        tip._add_child(child0)
        tip._add_child(child1)
        current_tips.extend([child0, child1])

    # assign tip labels as r{idx} for leaf nodes.
    tree = ToyTree(root)
    _assign_names(tree, random_names, rng)
    return tree


def unittree(
    ntips: int,
    treeheight: float = 1.0,
    random_names: bool = False,
    seed: Optional[int] = None,
) -> ToyTree:
    """Return an ultrametric tree where internal edges are all 1.

    A random tree is generated by iteratively splitting edges with
    equal probability. All internal edges are set to dist=1, and
    external edges are extended to make the tree ultrametric. The
    total tree height can be scaled to any arbitrary height, and
    other Nodes are scaled proportionately.

    Parameters
    ----------
    ntips: int
        The number of tips in the tree.
    treeheight: float
        All nodes in the tree will be scaled to retain relative node
        heights when scaling the total tree height to a new age.
    random_names: bool
        If True then names on the tips are randomized, else they are
        set to match with the node idx order.
    seed: Optional[int]
        An integer seed for the np random number generator.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=10, treeheight=1e6)
    """
    # seed generator
    rng = np.random.default_rng(seed)

    # generate random topology with N tips
    tree = rtree(ntips, seed=rng.integers(0, 2**31))

    # extend terminal edges to align, scale total height
    tree.mod.edges_scale_to_root_height(treeheight, inplace=True)
    tree.mod.edges_extend_tips_to_align(inplace=True)

    # randomize names
    _assign_names(tree, random_names, rng)
    return tree


def imbtree(
    ntips: int ,
    treeheight: float = 1.0,
    random_names: bool = False,
    seed: Optional[int] = None,
) -> ToyTree:
    """Return an imbalanced (ladder-like) tree topology.

    Parameters
    ----------
    ntips: int
        The number of tips in the tree.
    treeheight: float
        All nodes in the tree will be scaled to retain relative node
        heights when scaling the total tree height to a new age.
    random_names: bool
        If True then names on the tips are randomized, else they are
        set to match with the node idx order.
    seed: Optional[int]
        An integer seed for the np random number generator.

    Example
    -------
    >>> tree = toytree.rtree.imbtree(ntips=10, seed=123)
    """
    # generate random topology with N tips.
    root = Node()
    tip = root

    # add children n-1 times
    for _ in range(ntips - 1):
        child0 = Node(dist=1.)
        child1 = Node(dist=1.)
        tip._add_child(child0)
        tip._add_child(child1)
        tip = child0

    # will ladderize the tree and assign names and idxs
    tree = ToyTree(root)
    tree.mod.edges_extend_tips_to_align(inplace=True)
    tree.mod.edges_scale_to_root_height(treeheight, inplace=True)

    # randomize names
    _assign_names(tree, random_names, np.random.default_rng(seed))
    return tree


def baltree(
    ntips: int,
    treeheight: float = 1.0,
    random_names: bool = False,
    seed: Optional[int] = None,
) -> ToyTree:
    """Return a balanced tree topology.

    Raises an error in ntips is not an even number.

    Parameters
    ----------
    ntips: int
        The number of tips in the tree.
    treeheight: float
        All nodes in the tree will be scaled to retain relative node
        heights when scaling the total tree height to a new age.
    random_names: bool
        If True then names on the tips are randomized, else they are
        set to match with the node idx order.
    seed: Optional[int]
        An integer seed for the np random number generator.

    Example
    -------
    >>> tree = toytree.rtree.baltree(ntips=10, seed=123)
    """
    # in the construction algorithm below I store 'dist' values for
    # each node as the node distance from the root, to evenly create
    # splits. 'dists' are later overwritten by extend_tips_to_align.

    #check that ntips is even, if not raise an error
    if ntips%2 != 0:
        raise ToytreeError("Number of tips must be even to make a balanced tree!")

    # get a root Node and keep track of nancestors
    root = Node()

    # add children n-1 times. Visits leaf nodes
    curtips = 0
    while curtips < ntips - 1:

        # get node from smaller subtree
        parent = _get_small_child(root)

        # add children to the selected node and store
        child0 = Node(dist=1.)
        child1 = Node(dist=1.)

        # remove parent from dict
        parent._add_child(child0)
        parent._add_child(child1)
        curtips += 1

    # align tips, optionally set a to a height, add names, and return
    tree = ToyTree(root)
    tree.mod.edges_extend_tips_to_align(inplace=True)
    tree.mod.edges_scale_to_root_height(treeheight, inplace=True)
    _assign_names(tree, random_names, np.random.default_rng(seed))
    return tree


# def bdtree_stadler():
#     """More efficient method to simulate b-d trees (Stadler 2011).
#     """


def bdtree(
    ntips: int = 10,
    time: float = 4,
    b: float = 1,
    d: float = 0,
    stop: str = "taxa",
    seed: Optional[int] = None,
    retain_extinct: bool = False,
    random_names: bool = False,
    verbose: bool = False,
) -> ToyTree:
    """Return a parametric birth/death tree.

    The tree is generated by randomly sampling birth or death events
    starting from a single ancestor until a stopping criterion is
    reached. If the tree goes extinct it is restarted.

    The waiting time to the next speciation event is exponential with 
    parameter `b`; the waiting time to the next extinction event 
    is exponential with parameter `d`. Only one or the other will 
    be the next event. The waiting time to the next event (either type)
    is exponential with parameter b+d. The probability that an event
    is speciation is b/(b + d), extinction d/(b + d).

    Parameters
    ----------
    ntips: int
        Number of tips to generate for 'taxa' stopping criterion.
    time: float
        Amount of time to simulate for 'time' stopping criterion.
    b: float
        Birth rate per time unit
    d: death
        Death rate per time unit (d=0 produces Yule trees)
    stop: str
        Stopping critereon. Valid values are only 'taxa' or 'time'.
    seed: Optional[int]
        Random number generator seed.
    retain_extinct: bool
        (NotYetImplemented) Whether to retain internal nodes leading to
        extinct tips.
    random_names: bool
        Whether to randomize tip names or name them in order.
    verbose: bool
        Sets logging level to INFO to show statistics

    Examples
    --------
    >>> rtre = toytree.rtree.bdtree(ntips=10, b=1.0, d=0.5)
    """
    # require an appropriate option for stopping
    if stop not in ["taxa", "time"]:
        raise ToytreeError("stop must be either 'taxa' or 'time'")

    # random generator
    rng = np.random.default_rng(seed)

    taxa_stop = ntips
    time_stop = time

    # start from random tree (idxs will be re-assigned at end)
    root = Node(name="0")
    root.tdiv = 0

    # counters for extinctions, total events, and time
    resets = 0
    ext = 0
    evnts = 0
    time = 0

    # keep track of current leaf Nodes
    tips = [root]

    # continue until stop var
    while 1:

        # sample time until next event, increment t and events
        dtime = rng.exponential(1 / (len(tips) * (b + d)))
        time += dtime
        evnts += 1

        # sample a [0-1] to choose birth or death and sample a tip node
        rvar = rng.random()
        tip = rng.choice(tips)

        # event is a birth
        if rvar <= b / (b + d):
            # add child 1
            child1 = Node(name=f"{evnts}-1", dist=0)
            child1.tdiv = time
            tip._add_child(child1)
            # add child 2
            child2 = Node(f"{evnts}-2", dist=0)
            child2.tdiv = time
            tip._add_child(child2)
            # update tip list
            tips.extend([child1, child2])
            tips.remove(tip)

        # else event is extinction
        else:
            # get sisters
            sisters = tip.get_sisters()

            # drop the extinct tip
            tips.remove(tip)

            # sister exists and retains connect to unary parent
            if sisters:
                tip.up._remove_child(tip)

            # no sisters exist, so in addition to removing
            # this node we also remove any unary parent nodes
            # until we reach either the root, or a bipartition
            # since no descendants exist on this branch.
            else:
                while 1:
                    if tip.is_leaf() and not tip.is_root():
                        unary_node = tip
                        tip = tip.up
                        tip._remove_child(unary_node)
                    else:
                        break

            # if parent is None then reset
            if tip.up is None:
                resets += 1
                ext = 0
                evnts = 0
                root._dist = time = 0
                root._children = ()
                tips = [root]

            # advance extinction counter
            ext += 1

        # update branch lengths so all tips end at time=current
        for tip in tips:
            tip._dist += dtime

        # check stopping criterion
        if stop == "taxa":
            if len(tips) >= taxa_stop:
                break
        else:
            if time >= time_stop:
                break

    # log statistics
    if verbose:
        results = pd.Series({
            'time': time,
            'ntips': len(tips),
            'b': evnts - ext,
            'd': ext,
            'b/d': evnts / (evnts - ext),
            'resets': resets,
        })
        print(results)

    # if not retain_extinct then remove all internal unary nodes
    if not retain_extinct:
        # remove any unary nodes
        nodes = list(root._traverse_idxorder())
        for node in nodes[:-1]:
            if not node.is_leaf():
                if len(node.children) < 2:
                    node._delete(True, False)
        # make root the deepest node containing a split.
        while 1:
            if len(root.children) == 1:
                root = root.children[0]
            else:
                break

    # update coords and return
    # tre = tre.mod.ladderize()
    tree = ToyTree(root)
    _assign_names(tree, random_names, rng)
    return tree


def coaltree(
    k: int,
    N: int = 100,
    seed: Optional[int] = None,
    random_names: bool = False,
) -> ToyTree:
    """Return a random ToyTree generated under the n-coalescent.

    Waiting times between coal events under the n-coalescent are
    exponentially distributed with rate 4N / (k * k-1). A tree is
    constructed by randomly sampling waiting times from an exponential
    distribution at each value of k from k to 1, and randomly joining
    Nodes at each coalescent interval.

    Note
    ----
    The E[Tmrca] (expected root height) is 4N.

    Parameters
    ----------
    k: int
        The number of tips (e.g., gene copies) sampled at the present.
    N: int
        The diploid effective population size (Ne). Here this only
        acts only as a scalar multiplier of the branch lengths, but
        does not change their relative distribution, which is
        determined solely by the number of nodes between each
        coalescent event (i.e., the coalescent).
    seed: int or None
        A seed for the numpy random number generator.
    """
    # seed rng
    rng = np.random.default_rng(seed)

    # get waiting times
    kvals = np.arange(k, 1, -1)
    times = rng.exponential((4 * N) / (kvals * (kvals - 1)))

    # make tip Nodes for k samples
    nodes = {i: Node(name=str(i)) for i in range(k)}

    # iterate over coalescent time intervals
    for time in times:
        # increment the dist for all current Nodes
        for node in nodes:
            nodes[node]._dist += time

        # randomly sample two Nodes and connect to a new parent
        node0 = nodes.pop(rng.choice(tuple(nodes)))
        node1 = nodes.pop(rng.choice(tuple(nodes)))
        new_node = Node(str(k + 1))
        new_node._add_child(node0)
        new_node._add_child(node1)

        # advance name counter and update tips dict
        k = k + 1
        nodes[k] = new_node
    tree = ToyTree(new_node)
    _assign_names(tree, random_names, rng)
    return tree


def _get_small_child(node):
    """A recursive to return a Node from the smaller subclade of a
    tree, used for building balanced trees."""
    if not node.children:
        return node
    ldesc, rdesc = (len(i.get_leaves()) for i in node.children)
    if ldesc < rdesc:
        return _get_small_child(node.children[0])
    return _get_small_child(node.children[1])


# def coaltree(ntips, Ne, random_names=False, seed=None):
#     """Return a coalescent tree with ntips.
#
#     samples and waiting times ...
#     between coalescent events drawn from the kingman coalescent:
#     (4N)/(k*(k-1)), where N is effective population size (ne) and
#     k is sample size (ntips). Edge lengths on the tree are in
#     generations.
#     """
#     rng = np.random.default_rng(seed)

#     # convert units
#     coalunits = False
#     if not ne:
#         coalunits = True
#         ne = 10000

#     # build tree: generate N tips as separate Nodes then attach together
#     # at internal nodes drawn randomly from coalescent waiting times.
#     tips = [
#         toytree.tree().treenode.add_child(name=str(i))
#         for i in range(ntips)
#     ]
#     while len(tips) > 1:
#         rtree = toytree.tree()
#         tip1 = tips.pop(random.choice(range(len(tips))))
#         tip2 = tips.pop(random.choice(range(len(tips))))
#         kingman = (4. * ne) / float(ntips * (ntips - 1))
#         dist = random.expovariate(1. / kingman)
#         rtree.treenode.add_child(tip1, dist=tip2.height + dist)
#         rtree.treenode.add_child(tip2, dist=tip1.height + dist)
#         tips.append(rtree.treenode)

#     # build new tree from the newick string
#     self = toytree.tree(tips[0].write())
#     self.treenode.ladderize()

#     # make tree edges in units of 2N (then N doesn't matter!)
#     if coalunits:
#         for node in self.treenode.traverse():
#             node.dist /= (2. * ne)

#     # ensure tips are at zero (they sometime vary just slightly)
#     for node in self.treenode.traverse():
#         if node.is_leaf():
#             node.dist += node.height

#     # set tipnames to r{idx}
#     nidx = list(range(self.ntips))
#     if random_names:
#         random.shuffle(nidx)
#     for idx, node in enumerate(self):
#         if node.is_leaf():
#             node.name = "r{}".format(nidx[idx])

#     # decompose fills in internal node names and idx
#     self._coords.update()
#     return self


if __name__ == "__main__":

    TREE = rtree(10)
    print(TREE.get_tip_labels())

    TREE = unittree(10)
    print(TREE.get_tip_labels())

    TREE = bdtree(10, b=0.5, d=0.5, verbose=1)
    # TREE._draw_browser()
    # print(TREE.get_tip_labels())
    # print(TREE.treenode.draw_ascii())

    # TREE = bdtree(10)
    # print(TREE.get_tip_labels())

    # sim_trees = [
    #     toytree.rtree.rtree(10),
    #     toytree.rtree.baltree(10),
    #     toytree.rtree.imbtree(10),
    #     toytree.rtree.bdtree(10),
    #     toytree.rtree.unittree(10),
    # ]
    # print(sim_trees)


#!/usr/env/bin python

"""Tree distance metrics for measuring differences between trees.

Tree distance methods decompose trees into sets of bipartitions or
quartets, and use one of several methods for measuring differences
based on these sets, or additional features of the sets.

bipartition = {0, 1} | {2, 3, 4}
quartets = {0, 1} | {2, 3}; {0, 1} | {2, 4}; {0, 1} | {3, 4},

Authors
-------
- Deren Eaton, Scarlet Ming-sha Au

Note
----
If there was a real need for speed with tree distance functions
they could be easily sped up by adding jit decorators to various
set comparison functions, but this has not been done yet.

TODO: clean up references
References
----------
- https://en.wikipedia.org/wiki/Robinson%E2%80%93Foulds_metric
- Robinson, D.F. and Foulds, L.R. (1981) "Comparison of phylogenetic
  trees". Mathematical Biosciences. 53 (1–2): 131–147.
  doi:10.1016/0025-5564(81)90043-2.
- Smith, Martin R. (2020). "Information theoretic Generalized
  Robinson-Foulds metrics for comparing phylogenetic trees".
  Bioinformatics. 36 (20): 5007–5013. doi:10.1093/bioinformatics/btaa614.
- Mercè Llabrés, Francesc Rosselló, and Gabriel Valiente (2021)
  "The Generalized Robinson-Foulds Distance for Phylogenetic Trees"
  Journal of Computational Biology. http://doi.org/10.1089/cmb.2021.0342
- Sand, A, Holt, M., Johansen, J., Brodal, G., Mailund, T., Pedersen, C.,
  (2014) "tqDist: a library for computing the quartet and triplet
  distances between binary or general trees", Bioinformatics, Volume
  30, 14, 2079–2080, https://doi.org/10.1093/bioinformatics/btu157.
- Brodal 2013
- Estabrook 1985
- Holt 2014
- Sand 2014
- https://cran.r-project.org/web/packages/TreeDist/vignettes/Robinson-Foulds.html
- Meila M (2007). “Comparing clusterings-an information based
  distance.” _Journal of Multivariate Analysis_, *98*(5), 873-895.
  doi: 10.1016/j.jmva.2006.11.013 (URL:
  https://doi.org/10.1016/j.jmva.2006.11.013).
"""

from typing import Set, Callable, Union, Iterator, Tuple, Optional

from loguru import logger
import numpy as np
import pandas as pd
from toytree.distance._src.treedist_utils import (
    _get_split_phylo_info,
    get_trees_nye_dist,
    get_trees_matching_split_dist,
    get_trees_matching_split_info_dist,
    get_trees_shared_phylo_info_dist,
    get_trees_mutual_clust_info_dist,
    get_trees_shared_phylo_info_dist_from_biparts,
    get_trees_mutual_clust_info_dist_from_biparts,
)
from toytree import ToyTree
from toytree.core.apis import TreeDistanceAPI, add_subpackage_method
from toytree.utils import ToytreeError

logger = logger.bind(name="toytree")


TIPS_IDENTICAL = "Treedist methods require that trees share identical tip names."


# put functions here to have then exposed to 'distance' subpackage API
__all__ = [
    "get_treedist_rf",
    "get_treedist_rfi",
    "get_treedist_rfg_ms",
    "get_treedist_rfg_msi",
    "get_treedist_rfg_spi",
    "get_treedist_rfg_mci",
]


###################################################################
# Get distance metrics from bipartition sets and additional info
###################################################################

def _get_rf_distance(
    set1: Set, set2: Set, normalize: bool = True,
) -> float:
    """Return RF distance between two bipartition sets.

    The Robinson-Foulds distance is a normalized count of the
    bipartitions induced by one tree, but not the other tree, i.e.,
    it is the symmetric difference between two bipart sets divided
    by the total number of (internal) bipartitions in both sets.

    Parameters
    ----------
    set1: set
        A set of bipartitions in tree1 (e.g., tree1.iter_bipartitions)
    set2: set
        A set of bipartitions in tree2 (e.g., tree2.iter_bipartitions)
    normalize: bool
        Normalize distance score by the total number of splits in
        both trees (the max number of possible differences).
    """
    # get symmetric difference of bipartitions sets
    sym_diff = set1 ^ set2
    score = len(sym_diff)

    # normalize by the total number of internal edges.
    if normalize:
        return score / sum(len(i) for i in (set1, set2))
    return score


def _get_rf_distance_information_corrected(
    set1: Set, set2: Set, normalize: bool = True,
) -> float:
    """Return the information-corrected Robinson-Foulds distance (rfi).

    This distance measure is the sum of of the phylogenetic information
    of edges that are different between two trees, where information
    is calculated as the probability that a randomly sampled binary
    tree of the same size contains the split. Splits that contain
    less information (e.g., a cherry vs a deep split) are more likely
    to arise by chance, and thus contribute less to the metric.

    See Also
    --------
    - `get_treedist_rf`

    References
    ----------
    - Martin Smith: https://cran.r-project.org/web/packages/TreeDist/vignettes/information.html
    """
    # get total phylo info from the union of splits
    total_info = sum(_get_split_phylo_info(s) for s in set1 | set2)

    # get sum of phylo info on the shared splits
    shared = set1 & set2
    shared_info = sum(_get_split_phylo_info(s) for s in shared)

    # distance = (total phylo info) - (shared phylo info)
    # normalization = distance / (... options). The default normalizer
    # in rfi is sum of phyinfo in both trees independently, following
    # the default in TreeDist.
    if normalize:
        # default normalization (True)
        norm1 = sum(_get_split_phylo_info(s) for s in set1)
        norm2 = sum(_get_split_phylo_info(s) for s in set2)
        if normalize in ["sum", True]:
            return (total_info - shared_info) / sum((norm1, norm2))
        # other options (min, max, avg, )
        # elif normalize == "max":
        #     norm = max(norm1, norm2)
        # elif normalize == "avg":
        #     norm = (norm1 + norm2) / 2.
        # elif normalize == "min":
        #     norm = min(norm1, norm2)
    return total_info - shared_info


##############################################################
#
#  PUBLIC METHODS
#
##############################################################

@add_subpackage_method(TreeDistanceAPI)
def get_treedist_rf(
    tree1: ToyTree,
    tree2: ToyTree,
    normalize: Union[bool, int, Callable] = False,
) -> float:
    """Return the Robinson-Foulds (RF) distance between two trees.

    The Robinson-Foulds distance is a normalized count of the
    bipartitions induced by one tree, but not the other tree, i.e.,
    it is the symmetric difference between two bipart sets divided
    by the total number of (internal) bipartitions in both sets.
    Larger values indicate that two trees are more different.

    Parameters:
        tree1: ToyTree
            An input ToyTree to compare to tree2.
        tree2: ToyTree
            An input ToyTree to compare to tree1.
        normalize: bool
            Normalize distance score by the total number of splits in
            both trees (the max number of possible differences).

    Examples:
        >>> t0 = toytree.rtree.unittree(ntips=10, seed=123)
        >>> t1 = toytree.rtree.unittree(ntips=10, seed=321)
        >>> t0.distance.get_treedist_rf(t1, normalize=False)
        ...
        >>> t0.distance.get_treedist_rf(t1, normalize=True)
        ...

    References:
        1. Robinson, D.F. and Foulds, L.R. (1981) "Comparison of phylogenetic
        trees". Mathematical Biosciences. 53 (1–2): 131–147.
        doi:10.1016/0025-5564(81)90043-2.
    """
    assert set(tree1.get_tip_labels()) == set(tree2.get_tip_labels()), TIPS_IDENTICAL
    set1 = set(tree1.iter_bipartitions(type=frozenset, sort=True))
    set2 = set(tree2.iter_bipartitions(type=frozenset, sort=True))
    return _get_rf_distance(set1, set2, normalize=normalize)


@add_subpackage_method(TreeDistanceAPI)
def get_treedist_rfi(
    tree1: ToyTree,
    tree2: ToyTree,
    normalize: bool = False,
) -> float:
    """Return the information-corrected Robinson-Foulds distance (rfi).

    This distance measure is the sum of of the phylogenetic information
    of edges that are different between two trees, where information
    is calculated as the probability that a randomly sampled binary
    tree of the same size contains the split. Splits that contain less
    information (e.g., a cherry vs a deep split) are more likely to
    arise by chance, and thus contribute less to the metric.

    Parameters
    ----------
    tree1: ToyTree
        A ToyTree to compare to tree2.
    tree2: ToyTree
        A ToyTree to compare to tree1.
    normalize: bool
        Normalize score relative to the sum of phylogenetic info
        present in both subtrees.

    References
    ----------
    - Martin Smith: https://cran.r-project.org/web/packages/TreeDist/vignettes/information.html
    """
    assert set(tree1.get_tip_labels()) == set(tree2.get_tip_labels()), TIPS_IDENTICAL
    set1 = set(tree1.iter_bipartitions(type=frozenset, sort=True))
    set2 = set(tree2.iter_bipartitions(type=frozenset, sort=True))
    return _get_rf_distance_information_corrected(set1, set2, normalize)


@add_subpackage_method(TreeDistanceAPI)
def get_treedist_rfg_ms(
    tree1: ToyTree,
    tree2: ToyTree,
    normalize: bool = True,
) -> float:
    """Return the Matching Split Distance.

    Parameters
    ----------
    tree1: ToyTree
        A ToyTree to compare to tree2.
    tree2: ToyTree
        A ToyTree to compare to tree1.
    normalize: bool
        Normalize score relative to the sum of phylogenetic info
        present in both subtrees.

    References
    ----------
    - Bogdanowicz & Giaro (2012)
    """
    return get_trees_matching_split_dist(tree1, tree2, normalize)


@add_subpackage_method(TreeDistanceAPI)
def get_treedist_rfg_msi(
    tree1: ToyTree,
    tree2: ToyTree,
    normalize: bool = True,
) -> float:
    """Return the Matching Split Information Distance.

    Parameters
    ----------
    tree1: ToyTree
        A ToyTree to compare to tree2.
    tree2: ToyTree
        A ToyTree to compare to tree1.
    normalize: bool
        Normalize score relative to the sum of phylogenetic info
        present in both subtrees.

    References
    ----------
    - Bogdanowicz & Giaro (2012)
    - Martin Smith (2020)
    """
    return get_trees_mutual_clust_info_dist(tree1, tree2, normalize)


@add_subpackage_method(TreeDistanceAPI)
def get_treedist_rfg_spi(
    tree1: ToyTree,
    tree2: ToyTree,
    normalize: bool = False,
) -> float:
    """Return generalized Robinson-Foulds Distance between two trees
    based on Shared Phylogenetic Information (SPI).

    Generalized Robinson–Foulds distances calculate tree similarity by
    finding an optimal matching of splits between two trees, even
    among non-identical splits, on which to compute shared information.
    Here the 'shared phylogenetic information' metric is used to
    meausure the information that two trees hold in common, following
    Martin Smith (2020).

    Parameters
    ----------

    See Also
    --------

    References
    ----------
    - Smith, Martin R. (2020). "Information theoretic Generalized
      Robinson-Foulds metrics for comparing phylogenetic trees".
      Bioinformatics. 36 (20): 5007–5013. doi:10.1093/bioinformatics/btaa614.
    """
    assert set(tree1.get_tip_labels()) == set(tree2.get_tip_labels()), TIPS_IDENTICAL
    return get_trees_shared_phylo_info_dist(tree1, tree2, normalize)


@add_subpackage_method(TreeDistanceAPI)
def get_treedist_rfg_mci(
    tree1: ToyTree,
    tree2: ToyTree,
    normalize: bool = False,
) -> float:
    """Return generalized Robinson-Foulds Distance between two trees
    based on Mutual Clustering Information (MCI).

    Generalized Robinson–Foulds distances calculate tree similarity by
    finding an optimal matching of splits between two trees, even
    among non-identical splits, on which to compute shared information.
    Here the 'mutual clustering information' metric is used to
    meausure the information that two trees hold in common.

    This is the recommended metric for tree comparisons by Martin
    Smith (2020).

    Parameters
    ----------
    tree1: ToyTree
        A tree that will be compared to the second.
    tree2: ToyTree
        A tree that will be compared to the first.
    normalize: bool
        ...

    See Also
    --------

    References
    ----------
    - https://cran.r-project.org/web/packages/TreeDist/vignettes/Robinson-Foulds.html
    """
    assert set(tree1.get_tip_labels()) == set(tree2.get_tip_labels()), TIPS_IDENTICAL
    return get_trees_mutual_clust_info_dist(tree1, tree2, normalize)


@add_subpackage_method(TreeDistanceAPI)
def get_treedist_rfg_jac(
    tree1: ToyTree,
    tree2: ToyTree,
    k=1,
    allow_conflict: bool = True,
    normalize: bool = False,
) -> float:
    """Return generalized Robinson-Foulds Distance between two trees
    based on Nye Similarity metric.

    The Jaccard Robinson Foulds distance is a Generalized
    Robinson-Foulds metric based on the tree comparison method of
    Nye et al. (2006), and extended by Böcker et al. (2013).

    An optimal matching of bipartitions is found between two trees
    where pair scores represent the size of the largest split that is
    consistent with both them, normalized against the Jaccard index.

    Parameters
    ----------
    k: int
        An arbitrary exponent to which to raise the Jaccard Index. The
        Nye metric uses k=1; as k increases to infinity the metric
        converges on the standard RF metric.
    allow_conflict: bool
        If True conflicting splits that be paired, else they are given
        a score of zero.
    normalize: bool
        If True the score is normalized by the ...

    Notes
    -----

    See Also
    --------
    `toytree.distance.visualize_matching`

    Examples
    --------
    >>> tree1 = toytree.rtree.unittree(10, seed=123)
    >>> tree2 = toytree.rtree.unittree(10, seed=321)
    >>> tree1.distance.get_treedist_rfg_jac(tree2)   # ...
    >>> tree1.distance.get_treedist_rfg_jac(tree2, k=2)   # ...
    >>> tree1.distance.get_treedist_rfg_jac(tree2, k=2, allow_conflict=False)  # ...

    References
    ----------
    - Nye et al. (2006)
    - Böcker et al. 2013)
    - https://ms609.github.io/TreeDist/reference/JaccardRobinsonFoulds.html
    """
    raise NotImplementedError("TODO")


def get_treedist_kf_branch_score(
    tree1: ToyTree,
    tree2: ToyTree,
) -> float:
    """Return the Kune-Felsenstein metric...

    The Branch Score Distance of Kuhner and Felsenstein (1994) compares
    two trees using information of their branch lengths. It finds all
    bipartitions in the tree, and their branch lengths, as well as all
    possible alternative bipartitions that are not in the tree, which
    are assigned branch lenghts of zero.

    Reference
    ---------
    - Kuhner, M. K. and Felsenstein, J. (1994) Simulation comparison of
      phylogeny algorithms under equal and unequal evolutionary rates.
      Molecular Biology and Evolution, 11, 459–468.
    """
    utree = tree.unroot()
    iter_biparts = tree.iter_bipartitions(include_singleton_partitions=True)
    for node, bipart in zip(tree, iter_biparts):
        if node._up.is_root():
            "add dist to ..."
        print((node, node.up), node._dist, bipart)


def get_treedist_matrix(
    *trees: ToyTree,
    metric: Callable,
    normalize: bool = True,
    **kwargs,
) -> pd.DataFrame:
    """Return ...

    This is a generalization of `get_treedist_x` methods that
    arranges results for multiple tree comparisons into a dataframe.

    Parameters
    ----------

    Examples
    --------
    >>> tree1 = toytree.rtree.unittree(ntips=10, seed=123)
    >>> tree2 = toytree.rtree.unittree(ntips=10, seed=321)
    >>> ...
    """
    raise NotImplementedError("TODO")


##############################################################
#
#  NORMALIZATION
#
##############################################################

def _expected_variation(
    tree1: ToyTree,
    tree2: ToyTree,
    metric: str,
    nsamples: int = 1e+4,
    normalize: bool = False,
    seed: Optional[int] = None,
    **kwargs,
):
    """Return expected treedist metric between N random trees.

    For most tree distant metrics normalized values are expected to
    fall between 0-1, however, a distance of 1 will rarely be achieved
    since even two completely random trees are likely to still share
    some amount of information (splits, phylo info, clustering, etc.).
    Therefore, you can instead generate a null upper bound for the
    expected distance between random trees to use as a normalizer.

    Parameters
    ----------
    ...

    Examples
    --------
    >>> # get expected RFI distance between two N-tip trees.
    >>> tree1 = toytree.rtree.unittree(ntips=10, seed=123)
    >>> tree2 = toytree.rtree.unittree(ntips=10, seed=321)
    >>> expect = expected_variation(tree1, tree2, metric="rfg_spi")
    >>> print(expect)

    >>> # get Z-score to test if observation deviates from expectation
    >>> dist = tree1.distance.get_distance_rfi(tree2)
    >>> print(dist, (dist - expect.mean) / expect.std))
    """
    nsamples = int(nsamples)
    data = pd.Series(
        index=["estimate", "stdev", "stderr", "nsamples"],
        name=f"{metric}_distance",
        dtype=float,
    )

    # select function that takes bipart sets as input
    if metric == "rf":
        metric = _get_rf_distance
    elif metric == "rfi":
        metric = _get_rf_distance_information_corrected
    elif metric == "rfg_spi":
        metric = get_trees_shared_phylo_info_dist_from_biparts
    elif metric == "rfg_mcl":
        metric = get_trees_mutual_clust_info_dist_from_biparts
    else:
        raise ToytreeError("metric not recognized")

    # get bipartition table as ndarray, and array names
    biparts1 = set(tree1.iter_bipartitions())
    btable = tree2._get_bipartitions_table(include_singleton_partitions=False).values
    names = np.array(tree2.get_tip_labels())

    # for each sample assign randomized names to tree2 topology
    reps = np.zeros(shape=nsamples)
    rng = np.random.default_rng(seed)
    for idx in range(nsamples):
        # get splits when names are randomized
        biparts2 = set(_iter_random_biparts(btable, names, rng))
        reps[idx] = metric(biparts1, biparts2, normalize=normalize, **kwargs)

    data['estimate'] = reps.mean()
    data['stdev'] = reps.std()
    data['stderr'] = data.stdev / np.sqrt(nsamples)
    data['nsamples'] = nsamples
    return data


def _iter_random_biparts(
    btable,
    names,
    rng: Optional[np.random.Generator] = None,
) -> Iterator[Tuple[Tuple[str], Tuple[str]]]:
    """Return bipartitions for a topology w/ names randomized.

    This is used to calculate the expected variation in treedists for
    metrics that do not have an expected saturating distance, and so
    it is useful to ask how different two random trees are expected
    to be for a given ntips size.
    """
    np.random.shuffle(names)
    for idx in range(btable.shape[0]):
        mask0 = btable[idx].astype(bool)
        mask1 = np.invert(mask0)
        yield (tuple(names[mask0]), tuple(names[mask1]))


##############################################################
#
#  VALIDATION FUNCTIONS FOR COMPARING TO R LIBRARY
#
##############################################################

def _test_with_treedist_r(
    tree1: ToyTree,
    tree2: ToyTree,
    method: str = "rf",
    normalize: bool = True,
) -> float:
    """Return distance calculated with R's TreeDist library.

    This is used only for validation/testing internally only. Users
    are not expected to have R or treedist installed, and no
    checks are performed. This is not used in unittests.
    """
    import subprocess
    import tempfile

    # get TreeDist function
    method = method.lower()
    if method == "rf":
        func = "RobinsonFoulds"
    elif method == "rfi":
        func = "InfoRobinsonFoulds"
    elif method == "rfg_spi":
        func = "PhylogeneticInfoDistance"
    elif method == "rfg_mci":
        func = "ClusteringInfoDist"
    else:
        raise NotImplementedError("could do, but didn't yet..")

    # write R script to a tmpfile
    script = "\n".join([
        "library(TreeDist)",
        "t1 <- ape::read.tree(text='{}')".format(tree1.write(None, None, None)),
        "t2 <- ape::read.tree(text='{}')".format(tree2.write(None, None, None)),
        "dist <- {}(t1, t2, normalize={})".format(func, str(normalize).upper()),
        "print(dist)"
    ])

    with tempfile.NamedTemporaryFile(mode='w', encoding="utf-8") as tmp:
        tmp.write(script)
        tmp.flush()

        # run Rscript on the tempfile and catch stdout
        cmd = ["Rscript", "--vanilla", tmp.name]
        args = {'args': cmd, 'stdout': subprocess.PIPE, 'stderr': subprocess.STDOUT}
        with subprocess.Popen(**args, encoding='utf-8') as proc:
            out, _ = proc.communicate()
        # log the error
        try:
            out = float(out.split()[-1])
        except (ValueError, IndexError):
            logger.error(out)
    return out


def _validate(tree1, tree2):
    """Return dataframe with results for toytree and TreeDist"""
    algorithms = {
        "rf": get_treedist_rf,
        "rfi": get_treedist_rfi,
        "rfg_spi": get_treedist_rfg_spi,
        "rfg_mci": get_treedist_rfg_mci,
        # "qrt": get_treedist_qrt,
    }
    anames = list(algorithms) + [i + "-norm" for i in algorithms]
    data = pd.DataFrame(
        index=anames,
        columns=["toytree", "TreeDist"],
        data=0,
    )

    for algo, func in algorithms.items():
        for norm in (True, False):
            toy = func(tree1, tree2, normalize=norm)
            trd = _test_with_treedist_r(tree1, tree2, algo, norm)
            aname = algo + "-norm" if norm else algo
            data.loc[aname] = toy, trd
    return data


if __name__ == "__main__":

    import toytree
    t1 = toytree.rtree.baltree(10)
    t2 = toytree.rtree.imbtree(10)
    # print(_expected_variation(t1, t2, 'rf'))
    # print(_expected_variation(t1, t2, 'rfi'))

    # print("\n bal v imb")
    # print(_validate(t1, t2))

    t1 = toytree.tree("(1, (2, (3, (4, (5, (6, (7, 8)))))));")
    t2 = toytree.tree("(1, (2, (3, (4, (5, (7, (6, 8)))))));")
    t3 = toytree.tree("(1, (2, (3, (5, (4, (6, (7, 8)))))));")
    t4 = toytree.tree("(1, (2, (3, 4, 5, (6, (7, 8)))));")

    tree = toytree.tree("(((A:0.1,D:0.25):0.05,C:0.01):0.2,(B:0.3,E:0.8):0.2);")
    # get_treedist_kf_branch_score(tree)

    # print("\n t1-2")
    # print(_validate(t1, t2))

    # print("\n t1-3")
    # print(_validate(t1, t3))

    # print("\n t2-4")
    # print(_validate(t2, t4))

    # t1 = toytree.rtree.baltree(10, random_names=True)
    # t2 = toytree.rtree.imbtree(10, random_names=True)
    # print(get_treedist_rf(t1, t2, True))
    # print(_expected_variation(t1, t2, 'rf', normalize=True))
    # print(_expected_variation(t1, t2, 'rfi', normalize=True))
    # print(_expected_variation(t1, t2, 'rfg_spi', normalize=True))
    # print(_expected_variation(t1, t2, 'rfg_mcl', normalize=True))
    # print(_expected_variation(t1, t2, 'rfi'))



#!/usr/bin/env python

"""
Organization and API access of the pcm subpackage.

The modules in :mod:`toytree.pcm.src` contain Classes that are
exposed in user-facing factory functions written and documented
in :mod:`toytree.pcm.src.api`. The API docs are built from `api`,

Examples
--------
>>> toytree.pcm.get_vcv()
>>> toytree.pcm.get_qmatrix()

>>> toytree.pcm.sim_discrete_markov_data()
>>> toytree.pcm.sim_continuous_data()

>>> toytree.pcm.div_equal_splits()
>>> toytree.pcm.div_tip_level()
>>> toytree.pcm.div_tip_level()
"""

# from .discrete_markov_model_fit import simulate_discrete_markov_data



#!/usr/bin/env python

"""...

"""

from typing import Mapping, Any, Dict, Sequence
import numpy as np
from loguru import logger
from toytree.style import TreeStyle, SubStyle
from toytree.utils import ToytreeError

logger = logger.bind(name="toytree")


def check_arr(values: Sequence[Any], label: str, size: int, ctype: type) -> np.ndarray:
    """Return a Sequence as an ndarray checked for expected dtype and size.

    Raises an exception if values is 1 > len(values) > size, or it is
    not of the supported type.
    """
    arr = np.array(values)
    sizes = [size] if isinstance(size, int) else size
    if arr.size not in sizes:
        raise ToytreeError(
            f"'{label}' len mismatch error: len={arr.size} should be "
            f"in {sizes}.")
    if not isinstance(arr[0], ctype):
        raise ToytreeError(
            f"'{label}' type not supported. You entered {type(arr[0])}, "
            f"should be len={ctype}.")
    return arr


def substyle_dict_to_css_dict(style: Mapping[str, Any]) -> Dict[str, Any]:
    """Return dict with css style keys, e.g., 'font_size' -> 'font-size'."""
    new = {}
    for key, val in style.items():
        if "_" in key:
            if key == "anchor_shift":
                css_key = "-toyplot-anchor-shift"
                new[css_key] = style[key]
            else:
                css_key = key.replace("_", "-")
                new[css_key] = style[key]
        else:
            new[key] = val
    return new


def tree_style_to_css_dict(style: TreeStyle) -> Mapping[str, Any]:
    """Return dict w/ css keys from a TreeStyle.

    """
    style = style.__dict__
    for key, val in style.items():
        if isinstance(val, SubStyle):
            style[key] = substyle_dict_to_css_dict(val.__dict__)
    return style


if __name__ == "__main__":

    import toytree
    from toytree.style import validate_style

    tree = toytree.rtree.rtree(5)
    style = tree.style.copy()
    style = validate_style(tree, style)

    print(substyle_dict_to_css_dict(tree.style.node_style))#{"baseline_shift": 10}))
    # print(tree_style_to_css_dict(style))
    # tree_style_to_css_dict(style)



#!/usr/bin/env python

"""Newick parsing functions.

The core functions here are `parse_newick_string` and
`parse_newick_string_custom`. The latter can can take custom parsing
functions as input, making it very powerful and flexible, but a bit
complex for the average user; the `parse_newick_string` function
offers a simpler interface that should be sufficient for parsing
most newicks. These funcs are used internally in the generic
parsing function `toytree.tree`.

Supported formats
-----------------
- ((a,b),c);                           # topology only
- ((a:3,b:3):4,c:1);                   # w/ edge lengths
- ((a:3,b:3)100:4,c:1)100;             # w/ edge lengths and supports
- ((a:3,b:3)D:4,c:1)E;                 # w/ internal names
- ((a:3,b:3)D:4[&prob=100],c:1);       # w/ comment metadata on edges
- ((a:3,b:3)D[&prob=1]:4[&len=4],c:1); # w/ comments on edges and nodes!
- (str)str[Any]:float[Any]             # most complex

References
----------
- https://gist.github.com/Ad115/34dfc6560b64779a40c1a929f560511b
"""

from typing import (
    Optional, List, Any, Sequence, Tuple, Callable, Dict, Iterator, Set)
import re
from functools import partial
from loguru import logger
import numpy as np
from toytree.core.node import Node
from toytree.core.tree import ToyTree
from toytree.utils import ToytreeError


logger = logger.bind(name="toytree")
PAIRS = {'(': '()', '[': '[]', '{': "{}"}
COLON_OUTSIDE_SQUARE_BRACKETS = re.compile(r'(?<!\[):|:(?!\])')
RESERVED_FEATURE_NAMES = ["idx", "height", "dist"]
NHX_ERROR = """\
Error parsing NHX (extended New Hampshire format) newick meta data.
  NHX format = "...[{{prefix}}{{key}}{{assign}}{{value}}{{delim}}{{key}}{{assign}}{{value}}...]"
  Your data snippet = "[{}]"
  parse args               your values           common entries
  -----------------------------------------------------------------
  feature_prefix           {:<8}              "", "&", "&&NXH:"
  feature_assignment       {:<8}              "=", ":"
  feature_delim            {:<8}              ",", ":"
Try modifying the parse args to match with your NHX meta data format.
"""


def _find_closing(string: str, start: int = 1, pair: Sequence[str] = '()') -> int:
    """Find index of next unmatched closing parenth from start position.

    Examples
    --------
    >>> newick = '((),())()'
    >>> _find_closing(newick, start=1, pair='()')  # 6
    >>> _find_closing(newick, start=2, pair='()')  # 2

    Parameters
    ----------
    string: str
        The newick string.
    start: int
        Index starting position in the newick string.
    pair: str
        Allows to specify different opening/closing str types as a pair.
    """
    # get the delimiters
    opening, closing = pair

    # get index of next ')'
    close_idx = string.index(closing, start)

    # if there is no more '(' then return the next ')' index
    if string.find(opening, start) == -1:
        return close_idx

    # get index of next '('
    next_open_idx = string.index(opening, start)

    # if the next '(' is past the next ')' return ')' index
    if close_idx < next_open_idx:
        return close_idx

    # else, find next closure from next opening index, repeat recursively
    skip = _find_closing(string, next_open_idx + 1, pair)
    return _find_closing(string, skip + 1, pair)


def _iter_split_non_nested(node_str: str, delim: str = ",") -> Iterator[str]:
    """Generator of substrings between delimiter in a string.

    The delimiter does not split string if it occurs within a nested
    set: (), [], {}. This is used to parse node children, and also to
    parse NHX meta data, while avoiding really poorly formatted meta
    data (e.g., includes the delimiter) from messing things up.
    """
    final = len(node_str)
    start = 0
    end = 0
    while 1:

        # end and yield last chunk if at end of string
        if end == final:
            child = node_str[start:end]
            if child:
                yield child
            break

        # select the next character
        char = node_str[end]

        # if char is in "([{" then skip to next ")]}" respectively
        if char in PAIRS:
            pair = PAIRS[char]
            end = _find_closing(node_str, end + 1, pair)

        # split on delimiter not inside of nest and yield previous chunk
        elif char == delim:
            child = node_str[start:end]
            yield child
            start = end = end + 1
            # end += 1

        # advance if a data char
        else:
            end += 1


def _walk_newick_subtrees(
    newick: str,
    aggregator: Callable[[str, Any, float, Any], Any] = None,
    dist_formatter: Callable[[str], float] = None,
    feat_formatter: Callable[[str], Any] = None
) -> Tuple[Any, Set[str]]:
    """Recursive func (private) for extracting nested newick subtrees.

    The return type depends on the aggregator function.
    """
    # split first node from the newick string (inner)root[x]:dist[y];
    # leaving 'inner' which contains this Node's children.
    inner, label, dist, nmeta, emeta = _node_str_to_data(newick)

    # iterator of substrings split on ',' to separate children.
    # ""                  no children
    # ","                 two children w/ no info
    # ",,"                multiple children w/ no info
    # "a,b"               children with names
    # "a:3,b:4"           children w/ names and dist"
    # "a[...]:100[...]"   child w/ '' and meta
    # "(c,d),e"           subtree and child(ren) w or w/o stuff
    subtrees = _iter_split_non_nested(inner, delim=",")

    # iterate over subtrees calling this function recursively on each
    # and storing their data to their parent Node's child_data list.
    edge_features = set()
    child_data = []
    for child in subtrees:
        args = (aggregator, dist_formatter, feat_formatter)
        child_obj, features = _walk_newick_subtrees(child, *args)
        child_data.append(child_obj)
        edge_features.update(features)
        # edge_features |= features

    # str to float format the dist values
    distance = 1. if dist is None else dist_formatter(dist)

    # str to dict format the meta features
    nmeta = {} if nmeta is None else feat_formatter(nmeta)
    emeta = {} if emeta is None else feat_formatter(emeta)
    edge_features.update(set(emeta))
    # edge_features |= set(emeta)

    # aggegator func converts nested data to dict, Node, or other format.
    all_meta = {**nmeta, **emeta}
    # all_meta = nmeta | emeta  # dict merge operator '|' only py3.9+
    return aggregator(label, child_data, distance, all_meta), edge_features


def _node_str_to_data(newick: str) -> Tuple[str, str, str, str, str]:
    """Return data from a Node string (inner, label, dist, nmeta, emeta)

    """
    # split this node from the rest of tree.
    if newick.startswith("("):
        eidx = _find_closing(newick)
        outer = newick[eidx + 1:]
        inner = newick[1: eidx]
    else:
        outer = newick
        inner = ""

    # extract info from Node and Edge if ":" occurs outside sq brackets
    # label:          -> (label, None)
    # label[anno]:    -> (label, anno)
    # [anno]:         -> ('', anno)
    parts = list(_iter_split_non_nested(outer, delim=":"))
    if len(parts) > 1:
        node, edge = parts
        label, nmeta = _split_label_and_meta(node)
        dist, emeta = _split_label_and_meta(edge)

    # special for awful mrbayes format root edge anno w/o ':' delimiter
    # [anno1][anno2]  -> ('', anno1)
    elif "][" in outer:
        node, edge = outer.split("][")
        label, nmeta = _split_label_and_meta(node + "]")
        dist, emeta = None, None

    # label           -> (label, None)
    # label[anno]     -> (label, anno)
    # ''              -> ('', None)
    # [anno]          -> ('', anno)
    else:
        label, nmeta = _split_label_and_meta(outer)
        dist, emeta = None, None
    return inner, label, dist, nmeta, emeta


def _split_label_and_meta(substring: str) -> Tuple[str, str]:
    """Return tuple with (label, meta) given an outer newick substring."""
    for i, j in enumerate(substring):
        if j in "[:":
            return substring[:i], substring[i + 1: -1]
    return substring, None


def distance_parser(dist: str) -> Optional[float]:
    """Default float distance parser and formatter."""
    return float(dist)


def meta_parser(
    features: str,
    prefix: str = "",
    delim: str = ",",
    assignment: str = "=",
) -> Dict[str, str]:
    """Return a dict with auto-formatted features.

    When using this 'auto' function the dtype of features will be
    inferred later by toytree inside of the 'parse_newick' function,
    based on the str values for each feature across all Nodes. For
    example, if all can converted to floats or ints they will be.

    This will return {} if no features present.

    Parameters
    ----------
    feats: str
        A feature string extracted from the newick data.
    prefix: str
        A string prefix in feats, often starting with '&'.
    delim: str
        A string delimiter between features in the feature string.
    assignment: str
        A string assignment character located between keys and values
        in the feature string.
    """
    # remove the prefix (e.g., &&NHX:) and check that prefix was not
    # entered wrongly.
    #   [x=3]
    #   [&x=3]       amp can appear within metadata
    #   [&&NHX:x=3]  colons can appear within metadata.
    feats = features.lstrip(prefix)

    # return empty dict if no meta data present
    if not feats:
        return {}

    # split on feature separator
    items = _iter_split_non_nested(feats, delim)

    # store map of feature names to values
    meta = {}

    # iterate over items splitting to key,val pairs on assignment
    # splitter and attempting to convert to float.
    try:
        for item in items:
            # accommodate simplest NHX as in the example of RaxmlBipartitionsBranchLabels
            # format which has only [value], with no prefix, delim, or assignment
            if assignment not in item:
                key = "label"
                value = item
            # else parse it as you would normally expect, e.g., [...key=value...]
            else:
                key, value = item.split(assignment)

            # try to treat the value as numeric, else default to str
            try:
                meta[key] = float(value)
            except ValueError:
                meta[key] = str(value)
    except ValueError as exc:
        msg = NHX_ERROR.format(
            features,
            *(f"\"{i}\"" for i in (prefix, assignment, delim))
        )
        logger.error(msg)
        raise ToytreeError(msg) from exc
    return meta


def parse_newick_string_custom(
    newick: str,
    dist_formatter: Callable[[str], float] = None,
    feat_formatter: Callable[[str], Dict[str, Any]] = None,
    aggregator: Callable[[str, List[Node], float, Any], Node] = None,
    internal_labels: Optional[str] = None,
) -> Tuple[ToyTree, List[str]]:
    """Return a ToyTree from a newick string.

    Recursive function to build connected Nodes from nested data in
    newick format, and return as a ToyTree. Features parsed from the
    newick can be formatted with a custom formatter function, or using
    the default auto-formatting, which aims to infer the proper dtype
    based on the data.

    For high-level usage of this function users must write their own
    custom parsing functions. However, to provide a simpler option,
    most popular formats can be parsed using a builtin option from
    the toytree function `toytree.tree`.

    See Also
    --------
    `toytree.tree`, `toytree.io`

    Parameters
    ----------
    dist_formatter: Callable
        A custom function for converting edge length data into a float
        or int value to be stored as the .dist attribute of Nodes.
        e.g., `dist_formatter=lambda x: round(x / 100, 2)` to store
        as a rounded float with two floating points.
    feat_formatter: Callable
        A custom function for parsing newick string metadata to Node
        feature data. An example is NHX (extended newick) format, or
        similar formats like used by mrbayes. If no function here then
        the entire metadata is stored as a string as feature "feature".
        The function should return a Dict[str,Any].
        >>> def feat_parser(feats, prefix='&NHX:, delim=',', assign='='):
        >>>     feats = feats[len(prefix)]
        >>>     items = feats.split(delim)
        >>>     return dict(zip(items.split(assign)))
        See docs for more examples. The default feat_formatter is
        similar to above but tries to infer value types.
    aggregator: Callable
        A custom function that takes (name, children, dist, features)
        and returns a Node object. This is used to recursively build
        the Node objects from tuples of extracted newick data.
    internal_labels: str or None
        Feature type of internal labels. If None it is inferred to be
        either 'name' or 'support' based on numeric or string types
        being present.
    """
    # raise exception if semicolon is missing then strip it.
    newick = newick.strip()
    assert newick.endswith(";"), "newick string must end with ';'"
    newick = newick[:-1]

    # raise exception if parentheses are imbalanced
    if newick.count("(") != newick.count(")"):
        raise ToytreeError("Newick string parentheses are imbalanced")

    # select formatting functions
    if feat_formatter is None:
        feat_formatter = meta_parser
    if dist_formatter is None:
        dist_formatter = distance_parser
    if aggregator is None:
        aggregator = node_aggregator

    # build the connected Nodes from newick w/ features saved.
    args = (newick, aggregator, dist_formatter, feat_formatter)
    treenode, edge_features = _walk_newick_subtrees(*args)

    # set default root dist to 0 (Note: other Node's w/o dist default=1.)
    treenode._dist = 0.

    # convert connected Nodes to a ToyTree
    tree = ToyTree(treenode)

    # set edge features to ensure proper polarization if re-rooted
    tree.edge_features.update(edge_features)

    # infer whether labels on internal nodes are names or supports
    tree = _infer_internal_label_type(tree, internal_labels)

    # return the final tree
    return tree


# default aggregator in `parse_newick_string_custom`
def node_aggregator(
    label: str,
    children: List[Node],
    distance: float,
    features: Dict[str, Any],
) -> Node:
    """Aggregator function to connect Nodes and add features.

    Example of expected newick format: ((a:10,b:10)100:20, c:30);
    This tree contains 5 nodes, a, b, c, and two unlabeled nodes,
    the (a,b) ancestor, and the root. Their dists are 10, 10, 110,
    30, and default (0). The (a,b) ancestor has a label of '100',
    which could be interpreted as either a node str 'name', or as a
    node int 'support' value. This is a problem.
    """
    node = Node(name=label)
    node._dist = distance
    for child in children:
        node._add_child(child)

    # if any metadata annotations (e.g., [&x=3]) store as features
    for key, value in features.items():
        if key in RESERVED_FEATURE_NAMES:
            key = f"__{key}"
            # logger.warning(f"NHX feature name {key} is reserved and has been changed to __{key}")
        setattr(node, key, value)
    return node


def _dict_aggregator(label, children, distance, features):
    """Not Used, only for testing."""
    return dict(
        label=label,
        children=children,
        features=(distance, features)
    )


def _infer_internal_label_type(tree: ToyTree, internal_labels: Optional[str]) -> ToyTree:
    """Return a ToyTree with Node 'name' and 'support' updated.

    Check type of 'name' labels on internal Nodes. If all are numeric
    then save them as 'support' values instead of 'name' values.
    Changes are made to the Nodes in-place.
    """
    # if support set as numeric type
    if internal_labels == "support":
        for idx in range(tree.ntips, tree.nnodes):
            node = tree[idx]
            try:
                node.support = float(node.name)
            except ValueError:
                node.support = np.nan
            node.name = ""

    # if user entered a diff name then use that
    elif isinstance(internal_labels, str):
        for idx in range(tree.ntips, tree.nnodes):
            node = tree[idx]
            setattr(node, internal_labels, node.name)
            if internal_labels != "name":
                node.name = ""

    # infer types, any errors cause internal labels to be str names.
    # To save support values there must be a numeric label for every
    # internal non-root Node.
    else:  # elif internal_labels is None:
        try:
            # try to convert all internal node 'names' to floats (raises ValueError if str)
            names = [i.name for i in tree[tree.ntips:-1]]
            supports = [float(i.name) for i in tree[tree.ntips:-1]]

            # try to convert floats to ints if no floating points
            if all(i.is_integer() for i in supports):
                supports = [int(i) for i in supports]

            # store internal node values as 'support' and set 'name' to empty.
            for idx, inode in enumerate(tree[tree.ntips:-1]):
                inode.support = supports[idx]
                inode.name = ""

            # root Node is likely empty, but may have a name or even support
            # value. If so, we will try to store it numeric first, then string.
            tree.treenode.support = np.nan
            if not tree.treenode.name:
                tree.treenode.name = ""
            else:
                try:
                    tree.treenode.support = float(tree.treenode.name)
                    if tree.treenode.support.is_integer():
                        tree.treenode.support = int(tree.treenode.support)
                    tree.treenode.name = ""
                except ValueError:
                    pass

        # internal Node labels are inferred to be string name labels or other.
        except ValueError:
            if any(names):
                logger.info(
                    "empty or non-numeric node labels detected and set "
                    "as 'name' feature, not 'support'")
    return tree


def parse_newick_string(
    newick: str,
    feature_prefix: str = "&",
    feature_delim: str = ",",
    feature_assignment: str = "=",
    internal_labels: Optional[str] = None,
) -> ToyTree:
    """Return a ToyTree from a newick string.

    Recursive function to build connected Nodes from nested data in
    newick format, and return as a ToyTree. Features parsed from the
    newick can be formatted with a custom formatter function, or with
    the default auto-formatting, which aims to infer the proper dtype
    based on the data.

    For high-level usage of this function users must write their own
    custom parsing functions. However, to provide a simpler option,
    most popular formats can be parsed using a builtin option from
    the toytree function `toytree.tree`.

    See Also
    --------
    `toytree.tree`, `totyree.io`, `toytree.io.parse_newick_custom`

    Parameters
    ----------
    feature_prefix: str
        If additional features are stored in the newick as comments,
        they should be inside of square brackets. These blocks often
        start with a prefix character, such as "&" or "&&NHX". Check
        your file if you are unsure.
    feature_delim: str
        The character used to delimit features inside of a comment
        block. This is usually "," or ":".
    feature_assignment: str
        The character separating feature names and values in a comment
        block used for assignment. This is usually "=" or ":".
    internal_labels: str or None
        The feature that is present on internal node labels. If None
        then it will be inferred from the values present. Internal
        labels are usually either 'support' or 'name'. If only numeric
        values are present then it is parsed as 'support' floats,
        but this can overridden here if set `internal_labels='name'`.

    Examples
    --------
    >>> nwk1 = "((A,B),(C,D));"
    >>> tree1 = parse_newick_string(nwk1)
    >>> print(tree1.get_node_data())
    >>>
    >>> nwk2 = "((a:1,b:2)0.99:3,(c:1,d:1)0.90:3)0.66:1;"
    >>> tree2 = parse_newick_string(nwk2)
    >>> print(tree2.get_node_data())
    """
    # set default args to the feature parser (NHX)
    feat_formatter = partial(
        meta_parser,
        prefix=feature_prefix,
        delim=feature_delim,
        assignment=feature_assignment,
    )

    # parse tree using custom func with custom subfuncs
    tree = parse_newick_string_custom(
        newick=newick,
        feat_formatter=feat_formatter,
        dist_formatter=distance_parser,
        aggregator=node_aggregator,
        internal_labels=internal_labels,
    )
    return tree


def test1():
    NWK = "((a,b)Name[&&&x=3,z=0]:30[&&&length=3,y=4],c);"
    print(NWK)
    TREE = parse_newick_string_custom(NWK)
    print(TREE.get_node_data())


def test2():
    NWK = "((a,b)Name[&&&x=3,z=0]:30[&&&length=3,y=4],c[&&&height=10]);"
    print(NWK)
    TREE = parse_newick_string(NWK, feature_prefix="&&&")
    print(TREE.get_node_data())


def test3():
    import toytree
    NWK = """
(((ADH2:0.1[&&NHX:S=human:E=1.1.1.1],
ADH1:0.11[&&NHX:S=human:E=1.1.1.1]):0.05[&&NHX:S=Primates:E=1.1.1.1:D=Y:B=100],
ADHY:0.1[&&NHX:S=nematode:E=1.1.1.1],
ADHX:0.12[&&NHX:S=insect:E=1.1.1.1]):0.1[&&NHX:S=Metazoa:E=1.1.1.1:D=N],
(ADH4:0.09[&&NHX:S=yeast:E=1.1.1.1],ADH3:0.13[&&NHX:S=yeast:E=1.1.1.1],
ADH2:0.12[&&NHX:S=yeast:E=1.1.1.1],
ADH1:0.11[&&NHX:S=yeast:E=1.1.1.1]):0.1[&&NHX:S=Fungi])[&&NHX:E=1.1.1.1:D=N];
ADH1:0.11[&&NHX:S=yeast:E=1.1.1.1]):0.1[&&NHX:S=Fungi])[&&NHX:E=1.1.1.1:D=N];
"""
    # print(NWK)
    t = toytree.tree(NWK, feature_delim=":", feature_prefix="&&NHX:", feature_assignment="=")
    print(t.get_node_data())


def test4():
    import toytree
    NWK = "((C,D)1,(A,(B,X)3)2,E)R;"
    print(toytree.tree(NWK).get_node_data())
    print(toytree.tree(NWK, internal_labels='name').get_node_data())


if __name__ == "__main__":

    test2()
    # test3()
    # test4()
    # print(meta_parser("&A=100", prefix="&", delim="", assignment="="))
    # print(meta_parser("100", prefix=""))



#!/usr/bin/env python

"""Data transformations for formatting node values nicely.

# Tuple shortcut arg
style_arg=(feature, 2, 10)
"""

from typing import Sequence, Any, Optional, TypeVar
import numpy as np
from scipy.interpolate import interp1d
from pandas.api.types import is_numeric_dtype
from toytree.utils import ToytreeError

ToyTree = TypeVar("ToyTree")
NAN_NOT_ALLOWED = (
    "NaN values are not allowed in feature range mapping when nan_values=None")
ONLY_NUMERIC_ALLOWED = (
    "Only numeric dtypes can be used w/ tuple format (feature, ...)")


def get_range_mapped_feature(
    tree: ToyTree,
    feature: str,
    min_value: float = 5,
    max_value: float = 15,
    nan_value: Optional[float] = 0.,
    tips_only: bool = False
) -> np.ndarray:
    """Return an array of float values mapped to feature data in a tree.

    Parameters
    ----------
    tree: ToyTree
        A ToyTree to extract feature data from.
    feature: str
        Name of a feature to extract from one or more Nodes in the tree.
    min_value: float
        The minimum value (min of the range).
    max_value: float
        The maximum values (max of the range).
    nan_value: float or None
        A value to return for nan values. If None then nan values will
        raise a ValueError.
    tips_only: bool
        If True then data is only projected and returned for tip Nodes.
    """
    if tips_only:
        values = tree.get_tip_data(feature).values
    else:
        values = tree.get_node_data(feature).values
    if not is_numeric_dtype(values):
        raise ToytreeError(ONLY_NUMERIC_ALLOWED)
    return get_range_mapped_values(values, min_value, max_value, nan_value)


def get_range_mapped_values(
    values: Sequence[Any],
    min_value: float = 5,
    max_value: float = 15,
    nan_value: Optional[float] = 0.
) -> np.ndarray:
    """Return values mapped to a value range better for plotting.

    For example, we want to use some feature data such as 'dist' values
    to map Node sizes during plotting, but the dist values are in at
    values of 1/1000 or 1000, i.e., too small or too large for plotting.
    We can use this function to map them into a range more appropriate
    for plotting, such as 5-15 px, while maintaining their _relative_
    differences.

    Parameters
    ----------
    values: Sequence
        A sequence of numeric values. Non numerics will raise TypeError.
    min_value: float
        The minimum value (min of the range).
    max_value: float
        The maximum values (max of the range).
    nan_value: float or None
        A value to return for nan values. If None then nan values will
        raise a ValueError.

    Note
    ----
    This funtion uses `scipy.interpolate.interp1d` for interpolation.

    See Also
    ---------
    - toytree.color.get_color_mapped_feature

    Examples
    --------
    >>> tree = toytree.rtree.imbtree(8)

    >>> # get values from 'dist' feature in range 1-5
    >>> dists = tree.get_node_data("dist")
    >>> mdists = toytree.style.get_range_mapped_feature(dists, 1, 5)
    >>> tree.draw(edge_width=mdists)

    >>> # or, use shortcut to range map values when plotting
    >>> tree.draw(edge_width=('dist', (1, 5)))
    """
    assert isinstance(min_value, (float, int)), "min_values must be int or float"
    assert isinstance(max_value, (float, int)), "min_values must be int or float"

    # get mask of NaN values
    nan_mask = np.isnan(values)
    if all(nan_mask):
        if nan_value is None:
            raise ToytreeError(NAN_NOT_ALLOWED)
        return np.repeat(nan_value, len(values))

    # build interpolator from x data values to y requested range.
    mapper = interp1d(
        x=(np.nanmin(values), np.nanmax(values)),
        y=(min_value, max_value),
    )
    mvalues = np.array([mapper(i) for i in values])

    # handle nans in the data
    if nan_mask.sum():
        if nan_value is None:
            raise ToytreeError(NAN_NOT_ALLOWED)
        mvalues[nan_mask] = nan_value

    # return the range mapped values
    return mvalues


# def normalize_values(
#     values: Sequence[Any],
#     min_value: int = 2,
#     max_value: int = 12,
#     nbins: int = 10,
# ) -> np.ndarray:
#     """Distribute values into bins spaced at reasonable sizes for plotting.

#     This is used in tree_style='p' to automatically scale Ne values
#     to plot as edge widths. The input 'values' arg will usually be
#     a pandas.Series.

#     TODO: get_node_data dtype setting.., rename as discretize values?
#     """
#     # missing values are not allowed
#     if np.isnan(values).any():
#         raise ToytreeError(
#             f"missing values are not allowed:\n{values}.")

#     # make copy of original
#     ovals = deepcopy(values)

#     # if 6X min value is higher than max then add this
#     # as a fake value to scale more nicely
#     vals = list(values)
#     if min(vals) * 6 > max(vals):
#         vals.append(min(vals) * 6)

#     # sorted vals list
#     svals = sorted(vals)

#     # put vals into bins
#     bins = np.histogram(vals, bins=nbins)[0]

#     # convert binned vals to widths in 2-12
#     newvals = {}
#     sizes = np.linspace(min_value, max_value, nbins)
#     for idx, inbin in enumerate(bins):
#         for _ in range(inbin):
#             newvals[svals.pop(0)] = sizes[idx]
#     return np.array([newvals[i] for i in ovals])


if __name__ == "__main__":

    # import ipcoal
    import toytree

    # generate a random species tree with 10 tips and a crown age of 10M generations
    tree = toytree.rtree.unittree(10, treeheight=1e6, seed=123)
    # create a new tree copy with Ne values mapped to nodes
    vtree = tree.set_node_data(
        feature="Ne",
        data={i: 2e5 for i in (6, 7, 8, 9, 12, 15, 17)},
        default=1e4,
    )

    print(vtree.get_node_data())
    # vtree._draw_browser(ts='p', admixture_edges=[(0, 12, 0.5, {'stroke': 'red'}, "hello")]);



#!/usr/bin/env python

"""Phylogenetic community structure module.

This will be reorganized into a subpackage in the future, like this:
>>> tree = toytree.rtree.unittree(10)
>>> mat = phylocom.simulate_community(tree, size=100, ...)
>>> phylocom.get_community_metric(tree, mat, metric="pd", null=None)

References
----------
- https://pedrohbraga.github.io/CommunityPhylogenetics-Workshop/CommunityPhylogenetics-Workshop.html#within-assemblage-phylogenetic-structure
- https://phylodiversity.net/
- Cavender‐Bares, J. , Kozak, K. H., Fine, P. V. and Kembel, S. W.
(2009), The merging of community ecology and phylogenetic biology.
Ecology Letters, 12: 693-715. doi:10.1111/j.1461-0248.2009.01314.x

PSV and PGLMM methods
---------------------
- Ives and Helmus 2011
- Helmus M.R., Bland T.J., Williams C.K. & Ives A.R. (2007)
  Phylogenetic measures of biodiversity. American Naturalist,
  169, E68-E83
- TODO: phylogenetic entropy, Allen et al. 2009

Validation
----------
Following the methods in this paper
- https://link.springer.com/chapter/10.1007/978-3-662-43550-2_19
which cites Cadotte et al. for the approach, we can simulate
communities, compute N different statistics, and perform clustering
on the outputs to show how similar the results of different metrics
are.


For textbook:
- Faith: goal is to maximize conservation of features.
"""

from typing import TypeVar, Optional, Union
from numpy.typing import ArrayLike
import numpy as np
import pandas as pd
import toytree


ToyTree = TypeVar("ToyTree")


# TODO: simulate abundances as lognormally distributed?
def simulate_community_data(
    tree: ToyTree,
    scalar: float=0,
    size: int=1,
    seed: Optional[int]=None,
    ) -> pd.DataFrame:
    """Return a binary (nsites, nspecies) community data matrix.

    Simulate communities under phylogenetic attraction or repulsion.
    A VCV is generated from the tree; a scalar (c) is positive
    for phylogenetic attraction, or negative for repulsion. Phylo
    effects are modeled by a Cholesky decomposition of the VCV if
    scalar is positive, or of the inverse of VCV if negative. Random
    values are drawn from a normal distribution (R), and the
    prob of species occurrence is logistic with probability:

    $$ p = \frac{e^{cLR}}{(1 - e^{cLR})} $$

    A pandas.DataFrame is returned with N (size) sites where every
    site includes at least one observed species from the phylogeny.

    Parameters
    ----------
    tree: ToyTree
        A toytree with edge lengths from which a variance-covariance
        matrix will be calculated.
    scalar: float
        A constant scalar representing the strength of phylogenetic
        effects. Positive value is attraction, negative is repulsion.
    size: int
        Number of replicate communities (sites) to simulate.
    seed: int or None
        A seed for the numpy random number generator.
    """
    rng = np.random.default_rng(seed)
    vcv = toytree.pcm.get_vcv_matrix_from_tree(tree)
    vcv = np.linalg.inv(vcv) if scalar < 0 else vcv
    l_mat = np.linalg.cholesky(vcv)
    communities = []
    for _ in range(size):
        r_mat = rng.normal(size=vcv.shape[0])
        inner = np.exp(scalar * np.dot(l_mat, r_mat))
        probs = inner / (1 + inner)
        comm = rng.binomial(n=1, p=probs)
        communities.append(comm)
    return pd.DataFrame(columns=tree.get_tip_labels(), data=communities)

def get_community_metric(
    tree: ToyTree,
    matrix: ArrayLike,
    metric="pd",
    null=None,
    ) -> Union[float, np.ndarray]:
    """Return a community metric given a tree and matrix.

    Supported metrics are also available in individual functions
    with more detailed documentation in the `toytree.pcm.phylocom`
    subpackage. These include ...

    Parameters
    ----------
    tree: ToyTree
        A phylogenetic tree with the global species pool as tips.
    matrix: pd.DataFrame
        Community matrix (nsites, nspecies) composed of booleans
        or ints where 0/False indicates absence, and N/True indicates
        presence (if integer >1 it is interpreted as abundance for
        metrics that use abundance, otherwise simply as presence).
    metric: str
        A string matching the name of a supported community metric.
        See supported metrics.
    null: None or str
        A string matching the name of a supported null model. See
        supported null models.

    Examples
    --------
    >>> ...
    >>>    metric    null_mean    null_std   effect_size    name
    >>> 0    ...         ...          ...        ...        'MPD'
    """



if __name__ == "__main__":

    # fetch a test dataset ...
    TREE = toytree.rtree.unittree(10, treeheight=10, seed=123)
    MATRIX = simulate_community_data(TREE, scalar=10, size=10, seed=123)
    print(MATRIX)



#!/usr/bin/env python

"""A phylomaker-type tool that returns a pruned tree from a list of names.

As an example, the R tool "V.PhyloMaker2" takes ~42 hours to load an
existing large tree and prune it down to a list of >350K supplied taxa.

References
----------
Yi, J & Hong Q (2023) V.PhyloMaker2: An updated and enlarged R package
that can generate very large phylogenies for vascular plants.
Plant Diversity. https://doi.org/10.1016/j.pld.2022.05.005

Data Sources
------------
[name                           scope                   ntips                               nnodes      reference]
[----------------------------------------------------------------------------------------------------------------]
[World Plants (WP)              vascular plants          355,576                            ...          Hassler, Michael (1994 - 2024): World Plants. Synonymic Checklist and Distribution of the World Flora. Version 24.11; last update November 5th, 2024.]
[Leipzig Plant Catalogue (LCVP) vascular plants          351,180 (+6,160 hybrids)           ...          Freiberg M (2020). The Leipzig catalogue of vascular plants. German Centre for Integrative Biodiversity Research (iDiv) Halle-Jena-Leipzig. Checklist dataset https://doi.org/10.15468/9qxmn3 accessed via GBIF.org on 2024-11-25.]
[TreeFam (TF)                   animal gene trees        ...                                ...          Jue Ruan, Heng Li, et al. Nucleic Acids Research (2008) doi: 10.1093/nar/gkm1005]
[BirdTree (BT)                  birds                    9,993                              ...          Jetz, W., G. H. Thomas, J. B. Joy, K. Hartmann, and A. O. Mooers. 2012. The global diversity of birds in space and time. Nature 491:444-448    ]


Development Plan
-----------------
return dataframe w/ [tree-name, taxon-scope, ntips, nnodes, reference]
>>> toytree.infer.phylomaker.show_sources()
fetch newick of a big tree (from web) and write to disk.
>>> toytree.infer.phylomaker.fetch_newick(name)
fetch newick of a big tree (from web) and parse to toytree.
>>> toytree.infer.phylomaker.fetch_tree(name)
get a subtree as a ToyTree from a tree source and list of taxa.
this fetches and parses the full tree and caches a pickled copy
of the Toytree so it does not need to be re-parsed if re-run. It
logs warnings about bad taxon names, etc. and prunes the tree to
allow internal node name sampling such as genus, family.
>>> toytree.infer.phylomaker.get_subtree(name, list, cache=False)
"""


if __name__ == "__main__":
    pass



#!/usr/bin/env python

"""Enum submodule for finding or iterating over topologies from
a defined tree space.

IN DEVELOPMENT

TODO
----
- Yule tree probs
- get n-maximally probable tree

"""

import numpy as np
from toytree.core import ToyTree


def get_unlabeled_trees(ntips: int) -> int:
    r"""Return unlabaled tree and its probability.

    Note
    ----
    In contrast to labeled trees, all unlabeled trees for ntips are 
    *not* equiprobable, since there are some topology shapes that can
    be formed more than others by swapping around the labels on a tree.
    ...
    """


def get_the_n_maximally_probable_tree(ntips: int) -> ToyTree:
    """Return the topology most probable under Yule model for ntips.

    Reference
    ---------
    - ...
    """


def get_yule_tree_probability(tree: ToyTree) -> float:
    """Return the probability of ... under a Yule model.

    The probability that a specific labeled history is generated under
    a Yule model is ... equation 3+> of Brown...

    Utility
    -------
    Under the Yule model two (and only two) lineages are randomly
    joined at a time until all lineages have been merged into a tree.
    For nlineages there is thus a space of possible Yule trees, within
    which not all tree topologies are equally likely. This function
    returns the probabilitiy of the input tree topology under the Yule
    model, given ntips.

    Reference
    ---------
    - Harding EF (1971) The probabilities of rooted tree-shapes generated by random bifurcation. Adv Appl Prob 3: 44–77.
    - Brown JKM (1994) Probabilities of evolutionary trees. Syst Biol 43: 78–91.
    - Aldous DJ (2001) Stochastic models and descriptive statistics for phylogenetic trees, from Yule to today. Stat Sci 16: 23–34.
    - Steel M, McKenzie A (2001) Properties of phylogenetic trees generated by Yule-type speciation models. Math Biosci 170: 91–112.
    """


def get_probability_gene_tree_matches_species_tree(internal_branch: float) -> float:
    """...

    TODO: use in the textbook.

    References
    ----------
    - Pamilo and Nei (1988)
    """

    # a quartet (rooted three taxon) tree has no anomalies b/c
    λ = internal_branch

    # probability the gene tree matches the species tree
    prob_true = 1 - (2 / 3) * np.exp(-λ)

    # probability the gene tree matches other 2 possible trees
    prob_false = (1 / 3) * np.exp(-λ)

    # no matter how small the internal branch, prob_true always > prob_false
    return prob_true > prob_false


if __name__ == "__main__":
    pass



#!/usr/bin/env python

"""Educational resources submodule.

IDEAS
-----
- tree traversal animations
- wright_fisher: simulation, animation, and pair w/ tree visualizations
- markov: [todo] markov model visualization w/ animations.
- substitution: Markov model sympy demonstration tools.
- ...

"""

from toytree.learn.src.wrightfisher import wright_fisher_simulation



#!/usr/bin/env python

"""helper functions for tree parsing odd file formats.

tree parsing specifics
----------------------
- toytree.io.parse_tree_from_newick_file
- toytree.io.parse_tree_from_nexus_file
x- toytree.io.parse_tree_from_mb_file
x- toytree.io.parse_tree_from_beast3_file
x- toytree.io.parse_tree_from_bpp_file
x- toytree.io.parse_tree_from_superbpp_file
"""

from typing import Union, TypeVar, Optional, Union
import re
from pathlib import Path
from toytree.core import ToyTree
from toytree.io.src.parse import parse_tree, parse_tree_object
from toytree.utils import ToytreeError

# temporary
# ToyTree = TypeVar("ToyTree")
MultiTree = TypeVar("MultiTree")


def read_newick(
    path: Union[str, Path],
    feature_prefix: str = "&",
    feature_delim: str = ",",
    feature_assignment: str = "=",
    internal_labels: Optional[str] = None,
) -> Union[ToyTree, MultiTree]:
    """Return a ToyTree or MultiTree from a newick file.

    The file should contain one or more newick tree strings each on
    a separate line. This will parse the newick strings including
    those in extended-newick format with data stored in square-bracket
    comment blocks and return ToyTrees with data stored to Nodes.

    Compared to `toytree.tree` this function has additional options
    for specifying the format of feature comment blocks, and of the
    internal node label types (e.g., support values or names).

    See Also
    --------
    `toytree.tree`, `toytree.io.parse_newick_custom`

    Parameters
    ----------
    feature_prefix: str
        If additional features are stored in the newick as comments,
        they should be inside of square brackets. These blocks often
        start with a prefix character, such as "&" or "&&NHX". Check
        your file if you are unsure.
    feature_delim: str
        The character used to delimit features inside of a comment
        block. This is usually "," or ":".
    feature_assignment: str
        The character separating feature names and values in a comment
        block used for assignment. This is usually "=" or ":".
    internal_labels: str or None
        The feature that is present on internal node labels. If None
        then it will be inferred from the values present. Internal
        labels are usually either 'support' or 'name'. If only numeric
        values are present then it is parsed as 'support' floats,
        but this can overridden here if set `internal_labels='name'`.

    Examples
    --------
    >>> nwk1 = "((A,B),(C,D));"
    >>> tree1 = toytree.io.read_newick(nwk1)
    >>> print(tree1.get_node_data())
    >>>
    >>> nwk2 = "((a:1,b:2)0.99:3,(c:1,d:1)0.90:3)0.66:1;"
    >>> tree2 = toytree.io.read_newick(nwk2)
    >>> print(tree2.get_node_data())
    """
    kwargs = {
        "feature_prefix": feature_prefix,
        "feature_delim": feature_delim,
        "feature_assignment": feature_assignment,
        "internal_labels": internal_labels,
    }
    return parse_tree(path, **kwargs)


def read_nexus(path: Union[str, Path], **kwargs) -> ToyTree:
    """Return a ToyTree or MultiTree from a NEXUS file.

    The NEXUS file must contain a 'trees' block that includes one
    or more trees. This function takes the same arguments as
    `read_newick` for parsing extended-newick format of newick strings.
    It has additional arguments to parse other data from the NEXUS
    file, such as from the ... block (TODO)...
    """
    return parse_tree_object(path, **kwargs)


def read_mb_file(path: Union[str, Path], **kwargs) -> Union[ToyTree, MultiTree]:
    """Return a ToyTree from a mrbayes tree file.

    Mrbayes .trees files are NEXUS format with trees recorded
    using an extended newick format (NHX-like) with comment blocks
    that include both edge and node statistics separately. A trees
    file may contain many trees (a posterior distribution) or a single
    tree (e.g., a consensus tree).

    Internal node values can include ranges with "," inside, which
    requires additional comma-parsing that is automated by this
    function, as opposed to using `read_nexus`. TODO.

    Returns
    -------
    ToyTree or MultiTree
        The returned type depends on whether one or more trees in file.
    """
    return parse_tree_object(path, **kwargs)


def read_bpp_file(path: str, **kwargs):
    """Return a ToyTree or MultiTree from a BPP "00" tree file.

    This format has NHX data separately on nodes and edges, which is
    not typical other than perhaps for FigTree.

    Example
    -------
    >>> (a,(b,c))[]
    (6:0.000898,(5:0.000420,(3:0.000071,(2:0.000061,1:0.000061)[&height_95%_HPD={0.00004600,0.00007400},theta=0.0003823]:0.000010)[&height_95%_HPD={0.00005500,0.00008600},theta=0.0008018]:0.000349)[&height_95%_HPD={0.00033100,0.00050700},theta=0.0102051]:0.000478)[&height_95%_HPD={0.00078400,0.00101700},theta=0.0064148]
    """
    return parse_tree_object(path, **kwargs)


if __name__ == "__main__":

    # get a tree with metadata features
    import toytree
    TREE = toytree.rtree.imbtree(4, treeheight=3)

    # write/read newick
    # NWK = TREE.write()
    # TRE = toytree.io.read_newick(NWK)
    # print(TRE.get_node_data())

    # Example usage
    newick_string = "(c,(a,b)[&x=3,y=2]:0.1)[&x=4,y=5]:0.5;"

    # write/read mrbayes NEXUS
    # BPP1 = "/home/deren/Downloads/bpp-cacti-res/clades5-nloci1K-sample500K_r3.figtree.nex"
    # read_nexus(BPP1)



#!/usr/bin/env python

"""Parse tree data from flexible inputs to pass to a newick parser.

"""

from typing import Union, TypeVar, List, Tuple, Mapping
import re
from pathlib import Path
from loguru import logger
import requests

from toytree.core import ToyTree
from toytree.core.multitree import MultiTree
from toytree.io.src.newick import parse_newick_string
from toytree.io.src.nexus import get_newicks_and_translation_from_nexus
from toytree.io.src.utils import replace_whitespace

logger = logger.bind(name="toytree")

# for removing white_ space from newicks
# WHITE_SPACE = re.compile(r"[\n\r\t ]+")
ILLEGAL_NEWICK_CHARS = re.compile(r"[:;(),\[\]\t\n\r=]")

# PEP 484 recommend capitalizing alias names
Url = TypeVar("Url")


def parse_generic_to_str(data: Union[str, Url, Path]) -> str:
    """Return str data from input whether it is str, file, or url.

    The returned data string could be any format, it is not yet
    checked at this point. For example, newick or nexus; one tree
    or multiple trees; extended NHX or not.
    """
    # Path: read file and yield string. But if a str path then is
    # found differently further below in str parsing.
    if isinstance(data, Path):
        if data.exists():
            with open(data, 'r', encoding='utf-8') as indata:
                return indata.read()
        raise IOError(f"Path {data} does not exist.")

    # str: check if it is URI, then Path, then newick.
    if isinstance(data, str):
        # is there any scenario (nex?) not to strip by default?...
        data = data.strip()

        # empty string
        if ";" in data:
            return data

        # ...
        if not data:
            return "(0);"

        # newick always starts with "(" whereas a filepath and
        # URI can never start with this, so... easy enough.
        if data[0] in ("(", "#"):
            return data

        # check for URI (hack: doesn't support ftp://, etc.)
        if data.startswith("http"):
            response = requests.get(data)
            response.raise_for_status()
            return response.text

        # check if str is actually Path (fails if filename is large)
        if Path(data).exists():
            with open(data, 'r', encoding="utf-8") as indata:
                return indata.read()
        else:
            raise IOError(
                "Tree input appears to be a file path "
                f"but does not exist: '{data}'")

    # if entered as bytes convert to str and restart
    elif isinstance(data, bytes):
        data = data.decode()
        parse_generic_to_str(data)

    # not str or Path then raise TypeError
    raise TypeError(f"Error parsing unrecognized tree data input: {data}.")


def parse_data_from_str(strdata: str) -> Tuple[List[str], Mapping[int, str]]:
    """Return list of newicks and translation dict.
    """
    # if nexus then parse [nwks] from trees block
    if strdata[:6].upper() == "#NEXUS":
        nwks, tdict = get_newicks_and_translation_from_nexus(strdata)
        tdict = tdict
    # if newick then optionally split into multiple newicks
    else:
        strdata = replace_whitespace(strdata)
        nwks = strdata.split("\n")
        tdict = {}
    return nwks, tdict


def translate_node_names(tree: ToyTree, tdict: Mapping[int, str]) -> ToyTree:
    """Check valid and translate names using nexus translation dictionary."""
    if tdict:
        for node in tree:
            # replace disallowed characters in names with '_'
            if node.name:
                clean_name = ILLEGAL_NEWICK_CHARS.sub("_", tdict[node.name])
                node.name = clean_name
    return tree


def parse_tree(data: Union[str, Url, Path], **kwargs) -> ToyTree:
    """Return a ToyTree parsed from flexible input types.

    """
    strdata = parse_generic_to_str(data)
    nwks, tdict = parse_data_from_str(strdata)
    tree = parse_newick_string(nwks[0], **kwargs)
    if len(nwks) > 1:
        logger.warning(
            f"Data contains ({len(nwks)}) trees.\n"
            "Loading first using `toytree.tree`. Use `toytree.mtree` "
            "to instead load a MultiTree.")
    return translate_node_names(tree, tdict)


def parse_multitree(data: Union[str, Url, Path], **kwargs) -> ToyTree:
    """Return a MultiTree parsed from flexible input types.

    """
    strdata = parse_generic_to_str(data)
    nwks, tdict = parse_data_from_str(strdata)
    mtree = MultiTree([])
    for nwk in nwks:
        tree = parse_newick_string(nwk, **kwargs)
        translate_node_names(tree, tdict)
        mtree.treelist.append(tree)
    return mtree


def parse_tree_object(data: Union[str, Url, Path], **kwargs) -> Union[ToyTree, MultiTree]:
    """Return a ToyTree or MultiTree parsed from flexible input types.

    """
    strdata = parse_generic_to_str(data)
    nwks, tdict = parse_data_from_str(strdata)
    if len(nwks) > 1:
        mtree = MultiTree([])
        for nwk in nwks:
            tree = parse_newick_string(nwk, **kwargs)
            mtree.treelist.append(translate_node_names(tree, tdict))
        return mtree
    tree = parse_newick_string(nwks[0], **kwargs)
    return translate_node_names(tree, tdict)


if __name__ == "__main__":

    TEST = "/home/deren/Downloads/Clustal_Omega_Dec3.txt"
    TEST1 = "((a,b)c);"
    TEST2 = """(
    (a,b)X,c)
    ;
"""

    print(parse_tree_object(TEST))
    print(parse_tree_object(TEST1))
    print(parse_tree_object(TEST2))
    print(parse_tree_object("https://eaton-lab.org/data/Cyathophora.tre"))
    print(parse_tree_object("https://eaton-lab.org/data/densitree.nex"))

    TEST4 = """\
#NEXUS
begin trees;
    translate
           1       apple,
           2       blueberry,
           3       cantaloupe,
           4       durian,
           ;
    tree tree0 = [&U] ((1,2),(3,4));
    tree tree1 = [&U] ((1,2),(3,4));
end;
"""
    print(parse_tree_object(TEST4))

    TEST5 = """\
(((a:1,b:1):1,(d:1.5,e:1.5):0.5):1,c:3);
(((a:1,d:1):1,(b:1,e:1):1):1,c:3);
(((a:1.5,b:1.5):1,(d:1,e:1):1.5):1,c:3.5);
(((a:1.25,b:1.25):0.75,(d:1,e:1):1):1,c:3);
(((a:1,b:1):1,(d:1.5,e:1.5):0.5):1,c:3);
(((b:1,a:1):1,(d:1.5,e:1.5):0.5):2,c:4);
(((a:1.5,b:1.5):0.5,(d:1,e:1):1):1,c:3);
(((b:1.5,d:1.5):0.5,(a:1,e:1):1):1,c:3);
"""
    print(parse_tree_object(TEST5))
#     tool = TreeIOParser(TEST5)
#     trees = tool.parse_multitree_auto()
#     print(trees[0].get_tip_labels())



#!/usr/bin/env python

"""Default function for drawing MultiTrees as a grid of trees.

"""

from typing import Tuple, Optional, Sequence, Union, List, TypeVar
from toytree.core import Canvas, Cartesian, Mark
from toytree.drawing.src.setup_grid import Grid
from toytree.style import get_base_tree_style_by_name
from toytree.annotate import add_axes_scale_bar
import toytree
from loguru import logger

MultiTree = TypeVar("MultiTree")
logger = logger.bind(name="toytree")


def draw_multitree(
    mtree: MultiTree,
    shape: Tuple[int, int] = (1, 4),
    shared_axes: bool = False,
    idxs: Optional[Sequence[int]] = None,
    width: Optional[int] = None,
    height: Optional[int] = None,
    margin: Union[float, Tuple[int, int, int, int]] = None,
    **kwargs,
    ) -> Tuple[Canvas, Cartesian, List[Mark]]:
    """Return a toyplot drawing of a grid of ToyTrees.

    The grid spacing can be controlled with shape and margin
    options, and trees can each be on their own axis or on a
    set of shared axes (same max height dimension), which can be
    better for highlighting differences in heights.

    Parameters
    ----------
    shape: Tuple[int,int]
        A tuple of (nrows, ncolumns) for a tree grid drawing. The
        dimension of this matrix determines the number of trees
        that will be plotted.
    shared_axes (bool):
        If True then the 'height' dimension will be shared among
        all trees, otherwise each tree is scaled to fill the
        space in its grid cell.
    idxs (int):
        The indices of trees in treelist that you want to draw. By
        default the first ncols*nrows trees are drawn, but you can
        select the 10-14th tree by entering idxs=[10,11,12,13].
        The selected trees will be displayed in the grid.
    width (int):
        Width of the canvas
    height (int):
        Height of the canvas
    margin: Tuple[int,int,int,int]
        Spacing between subplots in the grid in pixel units. A
        single value or a tuple for top, right, bottom, left.
    **kwargs (dict):
        Any style arguments supported by .draw() in toytrees.

    Examples
    --------
    >>> trees = [toytree.rtree.unittree(10) for i in range(10)]
    >>> mtre = toytree.mtree(trees)
    >>> mtre.draw(shape=(2, 3), width=800, edge_widths=4)
    """
    # legacy support: warn user of old deprecated args.
    if kwargs.get("nrows") or kwargs.get("ncols"):
        raise DeprecationWarning(
            "nrows and ncols args deprecated. Use shape=(nrows, ncols)")

    # get indices of trees that will be drawn
    nrows = max(1, shape[0])
    ncols = max(1, shape[1])
    if idxs is None:
        tidx = range(0, min(nrows * ncols, len(mtree.treelist)))
    else:
        tidx = [idxs] if isinstance(idxs, int) else list(idxs)

    # get all or subset of ToyTrees within the idxs range
    try:
        treelist = [mtree.treelist[i] for i in tidx]
    except IndexError as err:
        logger.error(f"bad idxs argument: {err}")
        raise err

    # if fixed_order=True get the ladderized consensus tree tip order
    # as a List[str] and set that as the fixed_order arg.
    if kwargs.get("fixed_order") is True:
        fixed_order = (
            toytree.MultiTree(treelist)
            .get_consensus_tree()
            .get_tip_labels()
        )
        kwargs["fixed_order"] = fixed_order

    # if less than 4 trees reshape ncols,rows to use ntrees
    if len(treelist) < 4:
        if nrows > ncols:
            nrows = len(treelist)
            ncols = 1
        else:
            nrows = 1
            ncols = len(treelist)

    # get layout first from direct arg then from treestyle
    if "ts" in kwargs:
        layout = get_base_tree_style_by_name(kwargs.get("ts")).layout
    elif "tree_style" in kwargs:
        layout = get_base_tree_style_by_name(kwargs.get("ts")).layout
    else:
        layout = kwargs.get("layout", 'r')

    # get the canvas and axes that can fit the requested trees.
    padding = kwargs.get("padding", 10)
    scale_bar = kwargs.get("scale_bar", False)
    grid = Grid(nrows, ncols, width, height, layout, margin, padding, scale_bar)
    canvas = grid.canvas
    axes = grid.axes

    # default style
    if "tip_labels_style" in kwargs:
        if "-toyplot-anchor-shift" not in kwargs["tip_labels_style"]:
            kwargs["tip_labels_style"]["-toyplot-anchor-shift"] = "10px"
        if "font-size" not in kwargs["tip_labels_style"]:
            kwargs["font-size"] = "10px"
    else:
        kwargs["tip_labels_style"] = {
            "-toyplot-anchor-shift": "10px",
            "font-size": "10px",
        }

    # add toytree-Grid mark to the axes
    marks = []
    tmpargs = kwargs.copy()

    # get max tree height for shared_axes top
    ymax = max(abs(i.treenode.height) for i in treelist)

    # add ToyTree marks
    ncells = grid.nrows * grid.ncols
    for idx in range(ncells):

        # get the axis
        axes = grid.axes[idx]

        # check if an input arg that should be singular was entered as a
        # list of ncells, in which case it should be applied individually
        # to each tree as a sequence.
        # ...

        # add the mark
        if idx < len(treelist):
            _, _, mark = treelist[idx].draw(axes=axes, **tmpargs)

        # store the mark
        marks.append(mark)

    # mod style axes
    for idx in range(grid.nrows * grid.ncols):

        # HACK \/\/\/\/\/\/\/\/\/\/\/\
        if shared_axes:
            # grid.axes[idx].y.domain.max = ymax
            mark.width = canvas.width / ncols
            mark.height = canvas.height / nrows
            mark.scale_bar = kwargs.get("scale_bar", False)
            add_axes_scale_bar(treelist[idx], grid.axes[idx], ymax=ymax)
            # set_axes_ticks_style(ymax, grid.axes[idx], mark, only_inside=True)

            # add an invisible spacer point. This does a much
            # better job than setting ticks alone.
            if mark.layout == 'd':
                grid.axes[idx].scatterplot(
                    mark.xbaseline + treelist[idx].ntips / 2,
                    mark.ybaseline + ymax,
                    color="transparent",
                )
            elif mark.layout == 'u':
                grid.axes[idx].scatterplot(
                    mark.xbaseline + treelist[idx].ntips / 2,
                    mark.ybaseline - ymax,
                    color="transparent",
                )
            elif mark.layout == 'l':
                grid.axes[idx].scatterplot(
                    mark.xbaseline + ymax,
                    mark.ybaseline + treelist[idx].ntips / 2,
                    color="transparent",
                )
            elif mark.layout == 'r':
                grid.axes[idx].scatterplot(
                    mark.xbaseline - ymax,
                    mark.ybaseline + treelist[idx].ntips / 2,
                    color="transparent",
                )

        # axes off if not scale_bar
        if kwargs.get("scale_bar", False) is False:
            if mark.layout in 'du':
                grid.axes[idx].y.show = False
                grid.axes[idx].x.show = False
            else:
                grid.axes[idx].y.show = False
                grid.axes[idx].x.show = False

        if kwargs.get("label"):
            label = kwargs.get("label")
            if not isinstance(label, str):
                label = label[idx]
            grid.axes[idx].label.text = label

    # add mark to axes
    return canvas, grid.axes, marks    


if __name__ == "__main__":

    import toytree    
    trees = [toytree.rtree.unittree(5) for i in range(10)]
    mtree = toytree.mtree(trees)
    c, a, m = draw_multitree(mtree, shape=(2, 8))
    toytree.utils.show([c], tmpdir="~")



#!/usr/bin/env python

"""Distance-based tree inference.

"""

from typing import Tuple, Iterator, Union
from numpy.typing import ArrayLike
import numpy as np
import pandas as pd
import toytree


def infer_neighbor_joining_tree(data: ArrayLike) -> toytree.ToyTree:
    """Return a ToyTree inferred by neighbor-joining from a distance matrix.

    Neighbor-joining is a clustering algorithm for building trees from
    a distance matrix. It does not assume a clock, and is guaranteed
    to recover the true tree if the distances reflect the distances
    among samples on the true tree.

    Parameters
    ----------
    data: pd.DataFrame
        A symmetric DataFrame with distances measured between samples.

    Example
    -------
    >>> ...
    """
    # store names if provided, else use numeric range
    if hasattr(data, "index"):
        names = data.index
    else:
        names = np.arange(data.shape[0])

    # store a copy of dataframe as an array that will be modified.
    arr = np.array(data).astype(float)

    # dict to store Nodes, starting with tips.
    nodes = {i: toytree.Node(name=i) for i in names}

    # iterate generator function to get next pair of Nodes to join.
    for i, j, v_i, v_j in iter_nj_algorithm(arr):

        # get ordered Node names from the nodes dict.
        names = list(nodes.keys())
        # print(f"{names[i]}\t{names[j]}\t{v_i:.3f}\t{v_j:.3f}")

        # get Nodes with i, j names and pop from dict
        node_i = nodes.pop(names[i])
        node_j = nodes.pop(names[j])

        # create new ancestral Node, store it in nodes, and connect.
        if nodes:
            new_name = f"{node_i.name}-{node_j.name}"
            node_a = toytree.Node(name=new_name)
            nodes[new_name] = node_a

            # connect i and j to it with edge lengths v_i and v_j
            node_i._dist = v_i
            node_j._dist = v_j
            node_a._add_child(node_i)
            node_a._add_child(node_j)

        # connect final pair of Nodes
        else:
            node_i._dist = v_i
            node_j._add_child(node_i)

    # return final ancestor Node as root of a ToyTree
    return toytree.ToyTree(node_j)


def iter_nj_algorithm(arr: ArrayLike) -> Iterator[Tuple[int, int, float, float]]:
    """Generator function to yield node indices and branch lengths.

    Each iteration of the neighbor-joining algorithm finds the pair
    of samples with the shortest average distance to all other
    samples. This generator yields the indices (i, j) of the pair given
    an input 2-D distance array, and the branch lengths (v_i, v_j) of
    each of these to their parent node.
    """
    # iterate and reduce matrix until all Nodes are joined
    while 1:

        # get neighbor values (u_i)
        uvals = arr.sum(axis=0) / (arr.shape[0] - 2)

        # get c_arr as d_ij - u_i - u_j
        c_arr = arr - uvals - np.expand_dims(uvals, 1)

        # mask diagonal and get (i,j) index of min off-diagonal value
        mask = np.ones(c_arr.shape, dtype=bool)
        np.fill_diagonal(mask, False)
        idxs, jdxs = np.where(c_arr == c_arr[mask].min())
        i = idxs[0]
        j = jdxs[0]

        # get branch lengths from i,j to new internal Node
        v_i = 0.5 * arr[i, j] + 0.5 * (uvals[i] - uvals[j])
        v_j = 0.5 * arr[i, j] + 0.5 * (uvals[j] - uvals[i])

        # yield the new Node info (i, j, v_i, v_j)
        yield i, j, v_i, v_j

        # update arr to remove i and j and insert ij ancestor dists.
        new_dim = arr.shape[0] - 1
        new_arr = np.zeros(shape=(new_dim, new_dim))
        mask = np.ones(arr.shape[0], dtype=bool)
        mask[[i, j]] = False
        new_arr[:new_dim - 1, :][:, :new_dim - 1] = arr[mask, :][:, mask]
        new_arr[-1, :-1] = new_arr[:-1, -1] = (arr[i] + arr[j] - arr[i, j])[mask] / 2.
        arr = new_arr

        # if new arr size is 2 yield final pair and end.
        if new_dim == 2:
            yield 0, 1, arr[0, 1], arr[0, 1]
            break


if __name__ == "__main__":

    # example from Felsenstein
    DATA = pd.DataFrame(
        index=["dog", "bear", "raccoon", "weasel", "seal", "sea lion", "cat", "monkey"],
        columns=["dog", "bear", "raccoon", "weasel", "seal", "sea lion", "cat", "monkey"],
        data=np.array([
            [0, 32, 48, 51, 50, 48, 98, 148],
            [32, 0, 26, 34, 29, 33, 84, 136],
            [48, 26, 0, 42, 44, 44, 92, 152],
            [51, 34, 42, 0, 44, 38, 86, 142],
            [50, 29, 44, 44, 0, 24, 89, 142],
            [48, 33, 44, 38, 24, 0, 90, 142],
            [98, 84, 92, 86, 89, 90, 0, 148],
            [148, 136, 152, 142, 142, 142, 148, 0],
        ])
    )

    # run tree inference and draw it.
    TREE = infer_neighbor_joining_tree(DATA)

    # root on 'monkey' at ~1/3 of length.
    # TREE = TREE.root("monkey", root_dist=30)
    TREE = TREE.mod.root_on_midpoint()  # ("monkey", root_dist=30)

    # plot with .dist values shown
    DISTS = [f"{i:.2f}" for i in TREE.get_node_data('dist')]
    TREE._draw_browser(
        scale_bar=True,
        node_labels=DISTS,
        node_labels_style={"anchor-shift": -12, "baseline-shift": -10},
        node_mask=(0, 1, 1),
        node_sizes=5,
        tip_labels_style={"anchor-shift": 10},
        tip_labels_align=True,
        width=400,
        height=400,
        tmpdir="~",
    )



#!/usr/bin/env python

"""Functions for calculating diversification related statistics.

"""

from typing import Union, TypeVar
from itertools import combinations
import pandas as pd
import numpy as np
import toytree
from toytree.pcm.src.utils import calculate_posterior


ToyTree = TypeVar("ToyTree")
MultiTree = TypeVar("MultiTree")

__all__ = [
    "get_tip_level_diversification", 
    "get_equal_splits",
]


def _get_equal_splits(tree: ToyTree) -> pd.Series:
    """Return equal splits (ES) metric sensu Redding and Mooers 2006. 
    See :meth:`.get_equal_splits`
    """
    # store results
    data = pd.Series(name="ES", index=tree.get_tip_labels(), dtype=np.float64)

    # traverse up to root from each tip
    for node in tree:
        name = node.name
        divrate = 0
        j = 1
        while node.up:
            divrate += node.up.dist / (2 ** j)
            node = node.up
            j += 1
        data[name] = divrate
    return data    


def _get_tip_level_diversification(tree: ToyTree):
    """Return tip-level diversification rates sensu Jetz 2012. 
    See :meth:`.get_tip_level_diversification`
    """
    # ensure tree is a tree
    data = 1 / get_equal_splits(tree)
    data.name = "DR"
    return data


def get_equal_splits(
    trees: Union[ToyTree, MultiTree, str],
    njobs: int=1,
) -> pd.DataFrame:
    """Return equal splits (ES) metric sensu Redding and Mooers 2006.

    Parameters
    ----------
    trees: ToyTree, MultiTree, or newick file of one or more trees.
        One or more input trees.
    njobs: int
        Distribute N jobs in parallel using ProcessPoolExecutor.

    Returns
    -------
    pandas.DataFrame or pandas.Series
        If multiple trees a DataFrame with posterior statistics is
        returned, else a Series with stats for a single tree.

    Reference
    ---------
    TODO
    """
    if isinstance(trees, toytree.ToyTree):
        return _get_equal_splits(trees)
    return calculate_posterior(_get_equal_splits, trees, njobs)


def get_tip_level_diversification(
    trees: Union[ToyTree, MultiTree, str],
    njobs: int=1,
    **kwargs,
) -> pd.DataFrame:
    """Return tip-level diversification rates sensu Jetz 2012.

    Parameters
    ----------
    trees: ToyTree, MultiTree, or newick file of one or more trees.
        One or more input trees.
    njobs: int
        Distribute N jobs in parallel using ProcessPoolExecutor.
    **kwargs: dict
        A dictionary of arguments to the Callable function.

    Returns
    -------
    pandas.DataFrame or pandas.Series
        If multiple trees a DataFrame with posterior statistics is
        returned, else a Series with stats for a single tree.
    """
    if isinstance(trees, toytree.ToyTree):
        return _get_tip_level_diversification(trees, **kwargs)
    return calculate_posterior(
        _get_tip_level_diversification, trees, njobs, **kwargs)


def get_prob_dist_of_ntips(
    n_starting_tips: int=1,
    time: float=1,
    speciation_rate: float=0.5,
    extinction_rate: float=0.5,
    n_max_tips: int=500,
) -> np.ndarray:
    """Return probability distribution of a b-d model ending in ntips.

    References
    ----------
    - Harmon Chapter 10 (section 10.2)
    - Foote et al. (1999)
    """
    net_div = speciation_rate - extinction_rate
    rel_ext = extinction_rate / speciation_rate

    # the prob that any particular lineage has gone extinct before t
    def alpha(time):
        return (
            (rel_ext * ((np.exp(net_div) * time) - 1)) / 
            (np.exp(net_div) * time - rel_ext)
        )

    # ...
    def beta(time):
        return (
            (np.exp(net_div) * time - 1) /        
            (np.exp(net_div) * time - rel_ext)
        )

    # a simpler set of equations can be used when n_starting=1
    # p0t = alpha(0) ** n_starting_tips
    # pnt = (1 - alpha(t)) * (1 - beta(t)) * beta(t) ** ()

    # UNFINISHED...
    jvals = np.arange(1, min(n_starting_tips, n_max_tips))
    j_choose_n0 = sum(1 for i in combinations(range(n_starting_tips), j) if i)
    
    j_minus_one = sum(1 for i in combinations(range(n_max_tips - 1), j - 1) if i)
    aterm = alpha(time) ** (n_starting_tips - j)
    bterm = beta(time) ** (n_max_tips - j)
    sumprob += j_choose_n0 * j_minus_one * aterm * bterm * ((1 - alpha(time)) * (1 - beta(time)))**j
    pnt = None


def get_net_diversification_rate(
    ntips: int, 
    age: float, 
    stem: bool=True,
    relative_extinction_rate: float=0, 
) -> float:
    r"""
    
    $r_{stem} = \frac{ln(n)}{t_{stem}}$
    $r_{crown} = \frac{ln(n) - ln(2)}{t_{crown}}$

    Parameters
    ----------
    ntips: int
        The size of the clade.
    age: float
        The age of the clade.
    stem: bool
        True if age represents stem age, else False if crown age.
    relative_extinction_rate: float
        Estimate of (extinction rate / speciation rate).

    References
    ----------
    - Magallon and Sanderson (2001)
    - Harmon chapter 11
    """
    ext = relative_extinction_rate
    if stem:
        if relative_extinction_rate:
            return  np.log(ntips * (1 - ext) + ext) / age
        return np.log(ntips) / age
    if relative_extinction_rate:
        inner1 = (ntips * (1 - ext**2)) / 2
        inner2 = 2 * ext
        inner3a = (1 - ext)
        inner_inner = (ntips * ext**2) - (8 * ext) + (2 * ntips * ext) + ntips
        inner3b = np.sqrt(ntips * inner_inner)
        inner3 = (inner3a * inner3b) / 2
        return (np.log(inner1 + inner2 + inner3) - np.log(2)) / age
    return (np.log(ntips) - np.log(2)) / age


def get_birth_death_likelihood(
    tree: ToyTree,
    speciation_rate: float,
    extinction_rate: float,
    time_to_present: float,
    time_to_root: float,
    condition_on_tree_existence: bool=True,
) -> float:
    """...

    Example
    -------
    >>> # get edge lengths from a tree
    >>> edges = tree.get_node_data("dist")
    >>> get_birth_death_likelihood()

    References
    ----------
    - Harmon Chapter 11
    - Nee et al. (1994)
    - Maddison
    - Morlon et al. 2011
    - Stadler 2013a
    """
    if not extinction_rate:
        return (tree.ntips - 1) / sum(i.dist for i in tree)


    # final likelihood equation is Harmon 11.18 - 11.19
    lik = "..."
    # return lik / λ[1 − E(troot)]**2 


def fit_birth_death_model(
    tree: ToyTree,
):
    pass


if __name__ == "__main__":

    # tres = [toytree.rtree.unittree(10) for i in range(10)]
    # mtree = toytree.mtree(tres)
    # print(get_equal_splits(mtree, 2))
    # print(get_tip_level_diversification(mtree, 2))
    
    TREE = toytree.rtree.bdtree(ntips=100, b=0.5, d=0.5)
    MTREE = [TREE, TREE, TREE]
    print(get_tip_level_diversification(MTREE, njobs=10))


#!/usr/bin/env python

"""Infer a tree from a distance matrix using UPGMA clustering.

"""

import numpy as np
import pandas as pd
import toytree


def infer_upgma_tree(data: pd.DataFrame) -> toytree.ToyTree:
    """Return a ToyTree inferred by UPGMA from a distance matrix.

    Parameters
    ----------
    ...

    Examples
    --------
    >>> ...
    """
    # convert data to an array for faster processing.
    arr = np.array(data, dtype=float)

    # store tip Nodes, and add ntips attribute.
    nodes = {}
    for name in data.index:
        node = toytree.Node(name=name, dist=0)
        node._ntips = 1
        nodes[name] = node

    # loop to cluster and reduce matrix until only 1 sample left.
    while arr.shape[0] > 1:

        # get ordered names from nodes dict
        names = list(nodes.keys())

        # get index (i,j) of minvalue on off-diagonal.
        mask = np.ones(arr.shape, dtype=bool)
        np.fill_diagonal(mask, False)
        min_value = arr[mask].min()
        idxs = np.where(arr == min_value)[0]

        # create a new interior Node as ancestor of (i,j)
        new_name = f"internal-{arr.shape[0]}"
        new_node = toytree.Node(name=new_name, dist=0)
        nodes[new_name] = new_node

        # connect i,j to new Node with equal branch lengths
        for idx in idxs:
            name = names[idx]
            node = nodes[name]
            node._dist = (min_value / 2) - node._height
            new_node._add_child(node)

        # set height of new Node to be used by its parent to find dist.
        new_node._height = node._height + node._dist

        # count n leaves descended from new Node
        new_node._ntips = sum(i._ntips for i in new_node._children)

        # update the distance matrix to remove i,j and add new Node.
        new_dim = arr.shape[0] - 1
        new_arr = np.zeros(shape=(new_dim, new_dim))

        # index of rows to keep
        nidxs = [i for i in range(arr.shape[0]) if i not in idxs]

        # # fill kept rows into the new_arr
        new_arr[:new_dim - 1, :][:, :new_dim - 1] = arr[nidxs, :][:, nidxs]

        # remove i,j from nodes, and get ntips of each.
        nti = nodes.pop(names[idxs[0]])._ntips
        ntj = nodes.pop(names[idxs[1]])._ntips

        # get avg distance of new Node to every other Node.
        diks = arr[idxs[0], nidxs]
        djks = arr[idxs[1], nidxs]
        wdiks = (nti / (nti + ntj)) * diks
        wdjks = (ntj / (nti + ntj)) * djks
        avg_dists = wdiks + wdjks
        new_arr[-1, :-1] = avg_dists
        new_arr[:-1, -1] = avg_dists

        # advance internal node name counter
        arr = new_arr

    # return the top-level Node as a ToyTree
    return toytree.ToyTree(nodes[new_name])


if __name__ == "__main__":

    # example from Felsenstein
    DATA = pd.DataFrame(
        index=["dog", "bear", "raccoon", "weasel", "seal", "sea lion", "cat", "monkey"],
        columns=["dog", "bear", "raccoon", "weasel", "seal", "sea lion", "cat", "monkey"],
        data=np.array([
            [0, 32, 48, 51, 50, 48, 98, 148],
            [32, 0, 26, 34, 29, 33, 84, 136],
            [48, 26, 0, 42, 44, 44, 92, 152],
            [51, 34, 42, 0, 44, 38, 86, 142],
            [50, 29, 44, 44, 0, 24, 89, 142],
            [48, 33, 44, 38, 24, 0, 90, 142],
            [98, 84, 92, 86, 89, 90, 0, 148],
            [148, 136, 152, 142, 142, 142, 148, 0],
        ])
    )

    # show the distance matrix
    # print(DATA)
    TREE = infer_upgma_tree(DATA)
    DIST = [f"{i:.2f}" for i in TREE.get_node_data('dist')]
    TREE._draw_browser(
        ts='n',
        scale_bar=True, 
        node_labels=DIST,
        node_sizes=15, 
        node_mask=False,
        node_markers="r2x1",
        width=400, height=400,
        tip_labels_style={"-toyplot-anchor-shift": 20, }
    )
